# Copyright © 2025 UChicago Argonne, LLC All right reserved
# Full license accessible at https://github.com//AdvancedPhotonSource/pty-chi/blob/main/LICENSE

from typing import Optional
from .movie_utils import MoviesManager
from .mappings import movie_setting_types
from .settings import MovieFileTypes, MovieFileSettings
from ptychi.timing.timer_utils import timer
import ptychi.reconstructors.base as base

MOVIES_MANAGER = MoviesManager()
ENABLE_MOVIES = False


def toggle_movies(enable: bool):
    global ENABLE_MOVIES
    ENABLE_MOVIES = enable


def clear_movie_globals():
    global MOVIES_MANAGER
    MOVIES_MANAGER = MoviesManager()


def add_new_movie_builder(settings: movie_setting_types, folder: str, movie_name: str):
    global MOVIES_MANAGER
    if MOVIES_MANAGER is None:
        MOVIES_MANAGER = MoviesManager()
    MOVIES_MANAGER.add_movie_builder(settings, folder, movie_name)


@timer()
def update_movie_builders(reconstructor: "base.Reconstructor"):
    global MOVIES_MANAGER
    MOVIES_MANAGER.record_all_frames(reconstructor)


def create_all_movies(file_type: MovieFileTypes):
    global MOVIES_MANAGER
    MOVIES_MANAGER.create_all_movies(file_type)


def create_movie(
    movie_name: str,
    file_type: MovieFileTypes,
    movie_file_settings: Optional[MovieFileSettings] = None,
):
    if not check_if_movie_builder_exists(movie_name):
        return
    global MOVIES_MANAGER
    MOVIES_MANAGER.movie_builders[movie_name].create_movie(file_type, movie_file_settings)


def get_movie_builder_settings(movie_name: str) -> movie_setting_types:
    if not check_if_movie_builder_exists(movie_name):
        return
    global MOVIES_MANAGER
    return MOVIES_MANAGER.movie_builders[movie_name].settings


def reset_movie_builders():
    global MOVIES_MANAGER
    MOVIES_MANAGER.reset_all()


def delete_intermediate_movie_files():
    global MOVIES_MANAGER
    MOVIES_MANAGER.delete_intermediate_movie_files()


def check_if_movie_builder_exists(movie_name: str) -> bool:
    global MOVIES_MANAGER
    if movie_name not in MOVIES_MANAGER.movie_builders.keys():
        print(f"{movie_name} is not found")
        return False
    else:
        return True



# Copyright © 2025 UChicago Argonne, LLC All right reserved
# Full license accessible at https://github.com//AdvancedPhotonSource/pty-chi/blob/main/LICENSE

from typing import Tuple, Literal, Optional, Protocol
import math
import logging

import torch
from torch import Tensor
import torch.signal

import ptychi.maths as pmath
from ptychi.api.types import ComplexTensor, RealTensor
from ptychi.timing.timer_utils import timer, InlineTimer

class PlacePatchesProtocol(Protocol):
    def __call__(
        self, image: Tensor, positions: Tensor, patches: Tensor, op: Literal["add", "set"] = "add"
    ) -> Tensor: ...


class ExtractPatchesProtocol(Protocol):
    def __call__(self, image: Tensor, positions: Tensor, shape: Tuple[int, int]) -> Tensor: ...


logger = logging.getLogger(__name__)


@timer()
def batch_slice(image: Tensor, sy: Tensor, sx: Tensor, patch_size: Tuple[int, int]) -> Tensor:
    """
    Slice patches from an image at given window positions. The patch size is determined
    from the starting and ending coordinates in each direction, and is assumed to be
    the same for all patches.
    
    Parameters
    ----------
    image : Tensor
        A (H, W) tensor of the image.
    sy : Tensor
        A (N,) tensor of integers giving the starting y-coordinates of the patches.
    sx : Tensor
        A (N,) tensor of integers giving the starting x-coordinates of the patches.
    patch_size : tuple of int
        A tuple giving the patch shape in pixels.

    Returns
    -------
    Tensor
        A tensor of shape (N, h, w) containing the extracted patches.
    """
    h, w = image.shape[-2:]
    if (
        sy.min() < 0 
        or sy.max() + patch_size[0] > image.shape[-2] 
        or sx.min() < 0 
        or sx.max() + patch_size[1] > image.shape[-1]
    ):
        raise ValueError("Patch indices are out of bounds.")
    
    x = torch.arange(patch_size[1], device=sx.device)[None, :]
    y = torch.arange(patch_size[0], device=sy.device)[None, :]
    x = x.expand(len(sx), x.shape[1])
    y = y.expand(len(sy), y.shape[1])
    x = x + sx[:, None]
    y = y + sy[:, None]
    inds = (y * w).unsqueeze(-1) + x.unsqueeze(1)
    patches = image.view(-1)[inds.view(-1)]
    patches = patches.reshape(len(sy), patch_size[0], patch_size[1])
    return patches
    

@timer()
def batch_put(
    image: Tensor, 
    patches: Tensor, 
    sy: Tensor, 
    sx: Tensor, 
    op: Literal["add", "set"] = "add"
) -> Tensor:
    """
    Slice patches from an image at given window positions. The patch size is assumed 
    to be the same for all patches.
    
    Parameters
    ----------
    image : Tensor
        A (H, W) tensor of the buffer to place the patches into.
    patches : Tensor
        A (N, h, w) tensor of the patches.
    sy : Tensor
        A (N,) tensor of integers giving the starting y-coordinates of the patches.
    sx : Tensor
        A (N,) tensor of integers giving the starting x-coordinates of the patches.
    op : Literal["add", "set"]
        The operation to perform. "add" adds the patches to the image, 
        "set" sets the patches to the image replacing the existing values.
    
    Returns
    -------
    Tensor
        A tensor of shape (H, W) containing the image with patches added or set.
    """
    h, w = image.shape[-2:]
    if (
        sy.min() < 0 
        or sy.max() + patches.shape[-2] > image.shape[-2] 
        or sx.min() < 0 
        or sx.max() + patches.shape[-1] > image.shape[-1]
    ):
        raise ValueError("Patch indices are out of bounds.")
    
    patch_size = patches.shape[-2:]
    x = torch.arange(patch_size[1], device=sx.device)[None, :]
    y = torch.arange(patch_size[0], device=sy.device)[None, :]
    x = x.expand(len(sx), x.shape[1])
    y = y.expand(len(sy), y.shape[1])
    x = x + sx[:, None]
    y = y + sy[:, None]
    inds = (y * w).unsqueeze(-1) + x.unsqueeze(1)
    image = image.reshape(-1)
    
    try:
        patches_flattened = patches.view(-1)
    except RuntimeError:
        patches_flattened = patches.reshape(-1)
        
    if pmath.get_allow_nondeterministic_algorithms():
        # scatter_add_ and scatter_ are non-deterministic but faster.
        if op == "add":
            image.scatter_add_(0, inds.view(-1), patches_flattened)
        else:
            image.scatter_(0, inds.view(-1), patches_flattened)
    else:
        image.index_put_((inds.view(-1),), patches_flattened, accumulate=(op == "add"))
    return image.reshape(h, w)


@timer()
def extract_patches_integer(
    image: Tensor, positions: Tensor, shape: Tuple[int, int]
) -> Tensor:
    """
    Extract patches from 2D object, assuming the positions are at the
    exact center between pixels, so that no interpolation is needed. 
    If a patch's footprint goes outside the image,
    the image is padded with zeros to account for the missing pixels.

    Parameters
    ----------
    image : Tensor
        A (H, W) tensor giving the whole image.
    positions : Tensor
        A tensor of shape (N, 2) giving the center positions of the patches in pixels.
        The origin of the given positions are assumed to be the TOP LEFT corner of the image.
        For even patch shapes, the positions are assumed be at half pixels (i.e., positions % 1 == 0.5)
        so that the starting and ending coordinates of the patches are integers.
    shape : tuple of int
        A tuple giving the patch shape in pixels.

    Returns
    -------
    Tensor
        A tensor of shape (N, H, W) containing the extracted patches.
    """
    sys = (positions[:, 0] - (shape[0] - 1.0) / 2.0).round().int()
    sxs = (positions[:, 1] - (shape[1] - 1.0) / 2.0).round().int()
    
    pad_lengths = [
        max(-sxs.min(), 0),
        max(sxs.max() + shape[1] - image.shape[1], 0),
        max(-sys.min(), 0),
        max(sys.max() + shape[0] - image.shape[0], 0),
    ]
    if any(pad_lengths):
        logging.warning(
            f"Patch extractor has to pad the image by {[int(x) for x in pad_lengths]} "
            "(left, right, top, bottom) to avoid out-of-bounds errors. This should be "
            "avoided whenever possible. Consider using a larger object size."
        )
        image = torch.nn.functional.pad(image[None], pad_lengths, mode="replicate")[0]
        sys = sys + pad_lengths[2]
        sxs = sxs + pad_lengths[0]
    
    patches = batch_slice(image, sys, sxs, patch_size=shape)
    return patches


@timer()
def place_patches_integer(
    image: Tensor, 
    positions: Tensor, 
    patches: Tensor, 
    op: Literal["add", "set"] = "add"
) -> Tensor:
    """
    Place patches into a 2D object, assuming the positions are at the
    exact center between pixels, so that no interpolation is needed. 
    If a patch's footprint goes outside the image,
    the image is padded with zeros to account for the missing pixels.

    Parameters
    ----------
    image : Tensor
        The whole image to place the patches into.
    positions : Tensor
        A tensor of shape (N, 2) giving the center positions of the patches in pixels.
        The origin of the given positions are assumed to be the TOP LEFT corner of the image.
        For even patch shapes, the positions are assumed be at half pixels (i.e., positions % 1 == 0.5)
        so that the starting and ending coordinates of the patches are integers.
    patches : Tensor
        A (N, H, W) or (H, W) tensor of image patches.
    op : Literal["add", "set"]
        The operation to perform. "add" adds the patches to the image, 
        "set" sets the patches to the image replacing the existing values.

    Returns
    -------
    Tensor
        A tensor with the same shape as the object with patches added onto it.
    """
    shape = patches.shape[-2:]
    sys = (positions[:, 0] - (shape[0] - 1.0) / 2.0).round().int()
    sxs = (positions[:, 1] - (shape[1] - 1.0) / 2.0).round().int()
    
    pad_lengths = [
        max(-sxs.min(), 0),
        max(sxs.max() + shape[1] - image.shape[1], 0),
        max(-sys.min(), 0),
        max(sys.max() + shape[0] - image.shape[0], 0),
    ]
    if any(pad_lengths):
        logging.warning(
            f"Patch placer has to pad the image by {[int(x) for x in pad_lengths]} "
            "(left, right, top, bottom) to avoid out-of-bounds errors. This should be "
            "avoided whenever possible. Consider using a larger object size."
        )
        image = torch.nn.functional.pad(image, pad_lengths)
        sys = sys + pad_lengths[2]
        sxs = sxs + pad_lengths[0]
    
    image = batch_put(image, patches, sys, sxs, op=op)
    if any(pad_lengths):
        image = image[
            pad_lengths[2] : image.shape[0] - pad_lengths[3],
            pad_lengths[0] : image.shape[1] - pad_lengths[1],
        ]
    return image


@timer()
def extract_patches_fourier_shift(
    image: Tensor, positions: Tensor, shape: Tuple[int, int], pad: Optional[int] = 1
) -> Tensor:
    """
    Extract patches from 2D object. If a patch's footprint goes outside the image,
    the image is padded with zeros to account for the missing pixels.

    Parameters
    ----------
    image : Tensor
        The whole image.
    positions : Tensor
        A tensor of shape (N, 2) giving the center positions of the patches in pixels.
        The origin of the given positions are assumed to be the TOP LEFT corner of the image.
    shape : tuple of int
        A tuple giving the patch shape in pixels.
    pad : Optional[int]
        If given, patches with larger size than the intended size by this amount are cropped
        out from the patches before shifting.

    Returns
    -------
    Tensor
        A tensor of shape (N, H, W) containing the extracted patches.
    """
    # Floating point ranges over which interpolations should be done
    sys_float = positions[:, 0] - (shape[0] - 1.0) / 2.0
    sxs_float = positions[:, 1] - (shape[1] - 1.0) / 2.0

    # Crop one more pixel each side for Fourier shift
    sys = sys_float.floor().int() - pad
    eys = sys + shape[0] + 2 * pad
    sxs = sxs_float.floor().int() - pad
    exs = sxs + shape[1] + 2 * pad

    fractional_shifts = torch.stack([sys_float - sys - pad, sxs_float - sxs - pad], -1)

    pad_lengths = [
        max(-sxs.min(), 0),
        max(exs.max() - image.shape[1], 0),
        max(-sys.min(), 0),
        max(eys.max() - image.shape[0], 0),
    ]
    image = torch.nn.functional.pad(image, pad_lengths)
    sys = sys + pad_lengths[2]
    eys = eys + pad_lengths[2]
    sxs = sxs + pad_lengths[0]
    exs = exs + pad_lengths[0]

    patches = batch_slice(image, sys, sxs, patch_size=[shape[i] + 2 * pad for i in range(2)])

    # Apply Fourier shift to account for fractional shifts
    if not torch.allclose(fractional_shifts, torch.zeros_like(fractional_shifts), atol=1e-7):
        patches = fourier_shift(patches, -fractional_shifts)
    patches = patches[:, pad : patches.shape[-2] - pad, pad : patches.shape[-1] - pad]
    return patches


@timer()
def place_patches_fourier_shift(
    image: Tensor, 
    positions: Tensor, 
    patches: Tensor, 
    op: Literal["add", "set"] = "add",
    adjoint_mode: bool = True,
    pad: Optional[int] = 1,
) -> Tensor:
    """
    Place patches into a 2D object. If a patch's footprint goes outside the image,
    the image is padded with zeros to account for the missing pixels.

    Parameters
    ----------
    image : Tensor
        The whole image.
    positions : Tensor
        A tensor of shape (N, 2) giving the center positions of the patches in pixels.
        The origin of the given positions are assumed to be the TOP LEFT corner of the image.
    patches : Tensor
        A (N, H, W) or (H, W) tensor of image patches.
    op : Literal["add", "set"]
        The operation to perform. "add" adds the patches to the image, 
        "set" sets the patches to the image replacing the existing values.
    adjoint_mode : bool
        If True, this function performs the exact adjoint operation of `extract_patches_fourier_shift`.
        This means it will run the adjoint operation of every step of the extraction 
        function in reverse order: it first zero-pads the patches, shifts them back, 
        and puts them back into the image. Turn it on if this function is used in
        backpropagating the gradient. Note that due to the zero-padding, ripple
        artifacts may appear around the borders of each patch, so it is not suitable
        for placing patches that are not gradients. In that case, set this to False,
        and it will skip the zero-padding and crop the patches before placing them
        to remove Fourier shift wrap-arounds.
    pad : Optional[int]
        If given, patches are padded (or cropped) by this amount before shifting. 
        The actual operations depend on `adjoint_mode`: when `adjoint_mode` is True, 
        the patches are padded with zeros; otherwise, pathces are cropped by this amount
        after shifting to remove the wrap-around portions.

    Returns
    -------
    Tensor
        A tensor with the same shape as the object with patches added onto it.
    """
    # If the input is a single patch, add the third dimension
    # and expand it to the correct number of patches
    if len(patches.shape) == 2:
        patches = patches[None].expand(len(positions), -1, -1)

    shape = patches.shape[-2:]
    
    if adjoint_mode:
        patch_padding = pad
        patches = torch.nn.functional.pad(patches, [patch_padding] * 4)
    else:
        patch_padding = -pad

    sys_float = positions[:, 0] - (shape[0] - 1.0) / 2.0
    sxs_float = positions[:, 1] - (shape[1] - 1.0) / 2.0

    # Crop one more pixel each side for Fourier shift
    sys = sys_float.floor().int() - patch_padding
    eys = sys + shape[0] + 2 * patch_padding
    sxs = sxs_float.floor().int() - patch_padding
    exs = sxs + shape[1] + 2 * patch_padding

    fractional_shifts = torch.stack([sys_float - sys - patch_padding, sxs_float - sxs - patch_padding], -1)

    pad_lengths = [
        max(-sxs.min(), 0),
        max(exs.max() - image.shape[1], 0),
        max(-sys.min(), 0),
        max(eys.max() - image.shape[0], 0),
    ]
    image = torch.nn.functional.pad(image, pad_lengths)
    sys = sys + pad_lengths[2]
    eys = eys + pad_lengths[2]
    sxs = sxs + pad_lengths[0]
    exs = exs + pad_lengths[0]

    if not torch.allclose(fractional_shifts, torch.zeros_like(fractional_shifts), atol=1e-7):
        patches = fourier_shift(patches, fractional_shifts)
    if not adjoint_mode:
        patches = patches[
            :, 
            abs(patch_padding) : patches.shape[-2] - abs(patch_padding), 
            abs(patch_padding) : patches.shape[-1] - abs(patch_padding)
        ]

    inline_timer = InlineTimer("add or set patches on image")
    inline_timer.start()
    image = batch_put(image, patches, sys, sxs, op=op)
    inline_timer.end()

    # Undo padding
    image = image[
        pad_lengths[2] : image.shape[0] - pad_lengths[3],
        pad_lengths[0] : image.shape[1] - pad_lengths[1],
    ]
    return image


class ObjectPatchInterpolator:
    def __init__(self, object_: ComplexTensor, position_px: RealTensor, size: torch.Size) -> None:
        # top left corner of object support
        xmin = position_px[-1] - size[-1] / 2
        ymin = position_px[-2] - size[-2] / 2

        # whole components (pixel indexes)
        xmin_wh = xmin.int()
        ymin_wh = ymin.int()

        # fractional (subpixel) components
        xmin_fr = xmin - xmin_wh
        ymin_fr = ymin - ymin_wh

        # bottom right corner of object patch support
        xmax_wh = xmin_wh + size[-1] + 1
        ymax_wh = ymin_wh + size[-2] + 1

        # reused quantities
        xmin_fr_c = 1.0 - xmin_fr
        ymin_fr_c = 1.0 - ymin_fr

        # barycentric interpolant weights
        self._weight00 = ymin_fr_c * xmin_fr_c
        self._weight01 = ymin_fr_c * xmin_fr
        self._weight10 = ymin_fr * xmin_fr_c
        self._weight11 = ymin_fr * xmin_fr

        # extract patch support region from full object
        self._object_support = object_[ymin_wh:ymax_wh, xmin_wh:xmax_wh]

    @timer()
    def get_patch(self) -> ComplexTensor:
        """interpolate object support to extract patch"""
        object_patch = self._weight00 * self._object_support[:-1, :-1]
        object_patch += self._weight01 * self._object_support[:-1, 1:]
        object_patch += self._weight10 * self._object_support[1:, :-1]
        object_patch += self._weight11 * self._object_support[1:, 1:]
        return object_patch

    @timer()
    def update_patch(self, object_update: ComplexTensor) -> None:
        """add patch update to object support"""
        self._object_support[:-1, :-1] += self._weight00 * object_update
        self._object_support[:-1, 1:] += self._weight01 * object_update
        self._object_support[1:, :-1] += self._weight10 * object_update
        self._object_support[1:, 1:] += self._weight11 * object_update


@timer()
def extract_patches_bilinear_shift(
    image: Tensor,
    positions: Tensor,
    shape: Tuple[int, int],
    round_positions: bool = False,
    *args, **kwargs,
) -> Tensor:
    if round_positions:
        positions = torch.round(positions).to(int)

    obj_patches = torch.zeros((len(positions), *shape), dtype=image.dtype)
    for i in range(len(positions)):
        interpolator = ObjectPatchInterpolator(
            image,
            positions[i],
            shape,
        )
        obj_patches[i] = interpolator.get_patch()
    return obj_patches


@timer()
def place_patches_bilinear_shift(
    image: Tensor,
    positions: Tensor,
    patches: Tensor,
    op: Literal["add", "set"] = "add",
    round_positions: bool = False,
    *args, **kwargs,
) -> Tensor:
    if op == "set":
        raise NotImplementedError("\"set\" operation is not supported.")

    if round_positions:
        positions = torch.round(positions).to(int)

    for i in range(len(positions)):
        if patches.shape[0] == len(positions):
            patch_input = patches[i]
        elif len(patches.shape) == 2:
            patch_input = patches
        else:
            raise ValueError("Incorrect patch size.")

        interpolator = ObjectPatchInterpolator(
            image,
            positions[i],
            patch_input.shape,
        )
        interpolator.update_patch(patch_input)
    return image


@timer()
def shift_images(
    images: Tensor, 
    shifts: Tensor, 
    method: Literal["bilinear", "fourier"] = "bilinear",
    adjoint: bool = False,
    pad: Optional[int] = 0,
) -> Tensor:
    """Shift a batch of images by a given amount.
    
    Parameters
    ----------
    images : Tensor
        A [N, H, W] tensor of images.
    shifts : Tensor
        A [N, 2] tensor of shifts in pixels. Use the same shift values as the
        forward shift operation even if `adjoint` is True; do not flip the sign.
    method : Literal["bilinear", "fourier"]
        The method to use for shifting.
    adjoint : bool
        If True, the adjoint of the shift operation is performed. For Fourier
        shift, it shifts the image by `-shifts`. For bilinear shift, it computes
        the adjoint of the bilinear shift operation.
    pad : Optional[int]
        If not None, the image is padded with edge values by this amount before
        shifting.
    
    Returns
    -------
    Tensor
        Shifted images.
    """
    if pad is not None and pad > 0:
        images = torch.nn.functional.pad(images, (pad, pad, pad, pad), mode="replicate")
    if method == "bilinear":
        if adjoint:
            images = adjoint_bilinear_shift(images, shifts)
        else:
            images = bilinear_shift(images, shifts)
    elif method == "fourier":
        if adjoint:
            images = fourier_shift(images, -shifts)
        else:
            images = fourier_shift(images, shifts)
    else:
        raise ValueError(f"Invalid shift method: {method}")
    if pad is not None and pad > 0:
        images = images[..., pad:-pad, pad:-pad]
    return images


@timer()
def fourier_shift(images: Tensor, shifts: Tensor, strictly_preserve_zeros: bool = False) -> Tensor:
    """
    Apply Fourier shift to a batch of images.

    Parameters
    ----------
    images : Tensor
        A [N, H, W] tensor of images.
    shifts : Tensor
        A [N, 2] tensor of shifts in pixels.
    strictly_preserve_zeros : bool
        If True, mask of strictly zero pixels will be generated and shifted
        by the same amount. Pixels that have a non-zero value in the shifted
        mask will be set to zero in the shifted image. This preserves the zero
        pixels in the original image, preventing FFT from introducing small
        non-zero values due to machine precision.

    Returns
    -------
    Tensor
        Shifted images.
    """
    if strictly_preserve_zeros:
        zero_mask = images == 0
        zero_mask = zero_mask.float()
        zero_mask_shifted = fourier_shift(zero_mask, shifts, strictly_preserve_zeros=False)
    ft_images = pmath.fft2_precise(images)
    freq_y, freq_x = torch.meshgrid(
        torch.fft.fftfreq(images.shape[-2]), torch.fft.fftfreq(images.shape[-1]), indexing="ij"
    )
    freq_x = freq_x.to(ft_images.device)
    freq_y = freq_y.to(ft_images.device)
    freq_x = freq_x.repeat(images.shape[0], 1, 1)
    freq_y = freq_y.repeat(images.shape[0], 1, 1)
    mult = torch.exp(
        1j
        * -2
        * torch.pi
        * (freq_x * shifts[:, 1].view(-1, 1, 1) + freq_y * shifts[:, 0].view(-1, 1, 1))
    )
    ft_images = ft_images * mult
    shifted_images = pmath.ifft2_precise(ft_images)
    if not images.dtype.is_complex:
        shifted_images = shifted_images.real
    if strictly_preserve_zeros:
        shifted_images[zero_mask_shifted > 0] = 0
    return shifted_images


def bilinear_shift(images: Tensor, shifts: Tensor) -> Tensor:
    """
    Apply bilinear shift to a batch of images.
    
    Parameters
    ----------
    images : Tensor
        A [N, H, W] tensor of images.
    shifts : Tensor
        A [N, 2] tensor of shifts in pixels.

    Returns
    -------
    Tensor
        Shifted images.
    """
    if torch.allclose(shifts, torch.zeros_like(shifts)):
        return images
    y, x = torch.meshgrid(torch.arange(images.shape[-2]), torch.arange(images.shape[-1]), indexing="ij")
    y = y.to(images.device)
    x = x.to(images.device)
    y = y.repeat(images.shape[0], 1, 1)
    x = x.repeat(images.shape[0], 1, 1)
    
    # We want to sample the image at these positions.
    y = y - shifts[:, 0].view(-1, 1, 1)
    x = x - shifts[:, 1].view(-1, 1, 1)
    
    y_f = y.floor().int()
    y_c = (y + 1).floor().int()
    x_f = x.floor().int()
    x_c = (x + 1).floor().int()
    
    w_00 = (y_c - y) * (x_c - x)
    w_01 = (y_c - y) * (x - x_f)
    w_10 = (y - y_f) * (x_c - x)
    w_11 = (y - y_f) * (x - x_f)
    
    y_f = y_f.clip(0, images.shape[-2] - 1).view(-1)
    y_c = y_c.clip(0, images.shape[-2] - 1).view(-1)
    x_f = x_f.clip(0, images.shape[-1] - 1).view(-1)
    x_c = x_c.clip(0, images.shape[-1] - 1).view(-1)
    
    orig_shape = images.shape
    batch_inds = torch.arange(images.shape[0]).int().repeat_interleave(images.shape[-1] * images.shape[-2])
    images = \
        w_00.view(-1) * images[batch_inds, y_f, x_f] + \
        w_01.view(-1) * images[batch_inds, y_f, x_c] + \
        w_10.view(-1) * images[batch_inds, y_c, x_f] + \
        w_11.view(-1) * images[batch_inds, y_c, x_c]
    images = images.reshape(orig_shape)
    return images


def adjoint_bilinear_shift(images: Tensor, shifts: Tensor) -> Tensor:
    """
    Adjoint of bilinear shift.
    """
    if torch.allclose(shifts, torch.zeros_like(shifts)):
        return images
    y, x = torch.meshgrid(torch.arange(images.shape[-2]), torch.arange(images.shape[-1]), indexing="ij")
    y = y.to(images.device)
    x = x.to(images.device)
    y = y.repeat(images.shape[0], 1, 1)
    x = x.repeat(images.shape[0], 1, 1)
    
    # We want to sample the image at these positions.
    y = y - shifts[:, 0].view(-1, 1, 1)
    x = x - shifts[:, 1].view(-1, 1, 1)
    
    y_f = y.floor().int()
    y_c = (y + 1).floor().int()
    x_f = x.floor().int()
    x_c = (x + 1).floor().int()
    
    w_00 = (y_c - y) * (x_c - x)
    w_01 = (y_c - y) * (x - x_f)
    w_10 = (y - y_f) * (x_c - x)
    w_11 = (y - y_f) * (x - x_f)
    
    y_f = y_f.clip(0, images.shape[-2] - 1).view(-1)
    y_c = y_c.clip(0, images.shape[-2] - 1).view(-1)
    x_f = x_f.clip(0, images.shape[-1] - 1).view(-1)
    x_c = x_c.clip(0, images.shape[-1] - 1).view(-1)
    
    orig_shape = images.shape
    batch_inds = torch.arange(images.shape[0]).int().repeat_interleave(images.shape[-1] * images.shape[-2])
    
    adjoint = torch.zeros_like(images)
    adjoint.index_put_((batch_inds, y_f, x_f), (w_00 * images).view(-1), accumulate=True)
    adjoint.index_put_((batch_inds, y_f, x_c), (w_01 * images).view(-1), accumulate=True)
    adjoint.index_put_((batch_inds, y_c, x_f), (w_10 * images).view(-1), accumulate=True)
    adjoint.index_put_((batch_inds, y_c, x_c), (w_11 * images).view(-1), accumulate=True)
    adjoint = adjoint.reshape(orig_shape)
    return adjoint
    

@timer()
def nearest_neighbor_gradient(
    image: Tensor, direction: Literal["forward", "backward"], dim: Tuple[int, ...] = (0, 1)
) -> Tuple[Tensor, Tensor]:
    """
    Calculate the nearest neighbor gradient of a 2D image.

    Parameters
    ----------
    image : Tensor
        a (... H, W) tensor of images.
    direction : str
        'forward' or 'backward'.
    dim : tuple of int, optional
        Dimensions to calculate gradient. Default is (0, 1).

    Returns
    -------
    Tuple[Tensor, Tensor]
        A tuple of 2 images with the gradient in y and x directions.
    """
    if not hasattr(dim, "__len__"):
        dim = (dim,)
    grad_x = None
    grad_y = None
    if direction == "forward":
        if 1 in dim:
            grad_x = torch.concat([image[:, 1:], image[:, -1:]], dim=1) - image
        if 0 in dim:
            grad_y = torch.concat([image[1:, :], image[-1:, :]], dim=0) - image
    elif direction == "backward":
        if 1 in dim:
            grad_x = image - torch.concat([image[:, :1], image[:, :-1]], dim=1)
        if 0 in dim:
            grad_y = image - torch.concat([image[:1, :], image[:-1, :]], dim=0)
    else:
        raise ValueError("direction must be 'forward' or 'backward'")
    return grad_y, grad_x


@timer()
def gaussian_gradient(image: Tensor, sigma: float = 1.0, kernel_size=5) -> Tuple[Tensor, Tensor]:
    """
    Calculate the gradient of a 2D image with a Gaussian-derivative kernel.

    Parameters
    ----------
    image : tensor
        A (... H, W) tensor of images.
    sigma : float
        Sigma of the Gaussian.

    Returns
    -------
    Tuple[Tensor, Tensor]
        A tuple of 2 images with the gradient in y and x directions.
    """
    r = torch.arange(kernel_size) - (kernel_size - 1) / 2.0
    kernel = -r / (math.sqrt(2 * math.pi) * sigma**3) * torch.exp(-(r**2) / (2 * sigma**2))
    grad_y = convolve2d(image, kernel.view(-1, 1), padding="same", padding_mode="replicate")
    grad_x = convolve2d(image, kernel.view(1, -1), padding="same", padding_mode="replicate")

    # Gate the gradients
    grads = [grad_y, grad_x]
    for i, g in enumerate(grads):
        m = torch.logical_and(g.abs() < 1e-6, g.abs() != 0)
        if torch.count_nonzero(m) > 0:
            logger.debug("Gradient magnitudes between 0 and 1e-6 are set to 0.")
            g = g * torch.logical_not(m)
            grads[i] = g
    grad_y, grad_x = grads
    return grad_y, grad_x


@timer()
def fourier_gradient(image: Tensor) -> Tuple[Tensor, Tensor]:
    """
    Calculate the gradient of an image using Fourier differentiation
    theorem: `fft(df/dx) = 2 * pi * i * u * fft(f)`.

    Parameters
    ----------
    image : Tensor
        A (... H, W) tensor of images.

    Returns
    -------
    Tuple[Tensor, Tensor]
        The y and x gradients.
    """
    u, v = torch.fft.fftfreq(image.shape[-2]), torch.fft.fftfreq(image.shape[-1])
    u, v = torch.meshgrid(u, v, indexing="ij")
    grad_y = torch.fft.ifft(torch.fft.fft(image, dim=-2) * (2j * torch.pi) * u, dim=-2)
    grad_x = torch.fft.ifft(torch.fft.fft(image, dim=-1) * (2j * torch.pi) * v, dim=-1)
    return grad_y, grad_x


@timer()
def fourier_shift_gradient(image: Tensor, step: float = 0.5) -> Tuple[Tensor, Tensor]:
    """
    Calculate the gradient of an image using Fourier shift. 

    Parameters
    ----------
    image : Tensor
        A (... H, W) tensor of images.

    Returns
    -------
    Tuple[Tensor, Tensor]
        A tuple of (... H, W) tensor of gradients in y and x directions.
    """
    orig_shape = image.shape
    if image.ndim == 2:
        image = image.unsqueeze(0)
    
    pad = math.ceil(step) + 1
    image = torch.nn.functional.pad(image, (pad, pad, pad, pad), mode="replicate")
    
    sx1 = torch.tensor([[0, -step]], device=image.device).repeat(image.shape[0], 1)
    sx2 = torch.tensor([[0, step]], device=image.device).repeat(image.shape[0], 1)
    gx = fourier_shift(image, sx1) - fourier_shift(image, sx2)
    
    sy1 = torch.tensor([[-step, 0]], device=image.device).repeat(image.shape[0], 1)
    sy2 = torch.tensor([[step, 0]], device=image.device).repeat(image.shape[0], 1)
    gy = fourier_shift(image, sy1) - fourier_shift(image, sy2)
    
    if pad > 0:
        gy = gy[..., pad:-pad, pad:-pad]
        gx = gx[..., pad:-pad, pad:-pad]

    if len(orig_shape) == 2:
        gy = gy[0]
        gx = gx[0]
    return gy, gx


@timer()
def get_phase_gradient(
    img: Tensor,
    fourier_shift_step: float = 0,
    image_grad_method: Literal[
        "fourier_shift", "fourier_differentiation", "nearest"
    ] = "fourier_shift",
    eps: float = 1e-6,
) -> Tensor:
    """
    Get the gradient of the phase of a complex 2D image by first calculating
    the spatial gradient of the complex image, then taking the phase of the
    complex gradient -- i.e., it takes the phase of the gradient rather than
    the gradient of the phase. This avoids the sharp gradients due to phase
    wrapping when directly taking the gradient of the phase.

    Parameters
    ----------
    img : Tensor
        A [N, H, W] or [H, W] tensor giving a batch of images or a single image.
    step : float
        The finite-difference step size used to calculate the gradient, if
        the Fourier shift method is used.
    finite_diff_method : enums.ImageGradientMethods
        The method used to calculate the phase gradient.
            - "fourier_shift": Use Fourier shift to perform shift.
            - "nearest": Use nearest neighbor to perform shift.
            - "fourier_differentiation": Use Fourier differentiation.
    eps : float
        A stablizing constant.

    Returns
    -------
    Tuple[Tensor, Tensor]
        A tuple of 2 tensors with the gradient in y and x directions.
    """
    if fourier_shift_step <= 0 and image_grad_method == "fourier_shift":
        raise ValueError("Step must be positive.")

    if image_grad_method == "fourier_differentiation":
        gy, gx = fourier_gradient(img)
        gy = torch.imag(img.conj() * gy)
        gx = torch.imag(img.conj() * gx)
    else:
        # Use finite difference.
        if img.ndim == 2:
            img = img.unsqueeze(0)
        pad = int(math.ceil(fourier_shift_step)) + 1
        img = torch.nn.functional.pad(img, (pad, pad, pad, pad))

        sy1 = torch.tensor([[-fourier_shift_step, 0]], device=img.device).repeat(img.shape[0], 1)
        sy2 = torch.tensor([[fourier_shift_step, 0]], device=img.device).repeat(img.shape[0], 1)
        if image_grad_method == "fourier_shift":
            # If the image contains zero-valued pixels, Fourier shift can result in small
            # non-zero values that dangles around 0. This can cause the phase
            # of the shifted image to dangle between pi and -pi. In that case, use
            # `finite_diff_method="nearest" instead`, or use `step=1`.
            complex_prod = fourier_shift(img, sy1) * fourier_shift(img, sy2).conj()
        elif image_grad_method == "nearest":
            complex_prod = img * torch.concat([img[:, :1, :], img[:, :-1, :]], dim=1).conj()
        else:
            raise ValueError(f"Unknown finite-difference method: {image_grad_method}")
        complex_prod = torch.where(
            complex_prod.abs() < complex_prod.abs().max() * 1e-6, 0, complex_prod
        )
        gy = pmath.angle(complex_prod, eps=eps) / (2 * fourier_shift_step)
        gy = gy[0, pad:-pad, pad:-pad]

        sx1 = torch.tensor([[0, -fourier_shift_step]], device=img.device).repeat(img.shape[0], 1)
        sx2 = torch.tensor([[0, fourier_shift_step]], device=img.device).repeat(img.shape[0], 1)
        if image_grad_method == "fourier_shift":
            complex_prod = fourier_shift(img, sx1) * fourier_shift(img, sx2).conj()
        elif image_grad_method == "nearest":
            complex_prod = img * torch.concat([img[:, :, :1], img[:, :, :-1]], dim=2).conj()
        complex_prod = torch.where(
            complex_prod.abs() < complex_prod.abs().max() * 1e-6, 0, complex_prod
        )
        gx = pmath.angle(complex_prod, eps=eps) / (2 * fourier_shift_step)
        gx = gx[0, pad:-pad, pad:-pad]
    return gy, gx


@timer()
def integrate_image_2d_fourier(grad_y: Tensor, grad_x: Tensor) -> Tensor:
    """
    Integrate an image with the gradient in y and x directions using Fourier
    differentiation.

    Parameters
    ----------
    grad_y, grad_x: Tensor
        A (H, W) tensor of gradients in y or x directions.

    Returns
    -------
    Tensor
        The integrated image.
    """
    shape = grad_y.shape
    f = pmath.fft2_precise(grad_x + 1j * grad_y)
    y, x = torch.fft.fftfreq(shape[0]), torch.fft.fftfreq(shape[1])

    # In PtychoShelves' get_img_int_2D.m, they set the numerator of r to be
    # exp(2j * pi * (x + y[:, None])) to shift it by 1 pixel. We should NOT
    # do this in order to get the same result as PtychoShelves.
    r = 1.0
    r = r / (2j * torch.pi * (x + 1j * y[:, None]))
    r[0, 0] = 0
    integrated_image = f * r
    integrated_image = pmath.ifft2_precise(integrated_image)
    if not torch.is_complex(grad_x):
        integrated_image = integrated_image.real
    return integrated_image


@timer()
def integrate_image_2d_deconvolution(
    grad_y: Tensor,
    grad_x: Tensor,
    tf_y: Optional[Tensor] = None,
    tf_x: Optional[Tensor] = None,
    bc_center: float = 0,
) -> Tensor:
    """
    Integrate an image with the gradient in y and x directions by deconvolving
    the differentiation kernel, whose transfer function is assumed to be a
    ramp function.
    
    Adapted from Tripathi, A., McNulty, I., Munson, T., & Wild, S. M. (2016). 
    Single-view phase retrieval of an extended sample by exploiting edge detection 
    and sparsity. Optics Express, 24(21), 24719–24738. doi:10.1364/OE.24.024719

    Parameters
    ----------
    grad_y, grad_x: Tensor
        A (H, W) tensor of gradients in y or x directions.
    tf_y, tf_x: Tensor
        A (H, W) tensor of transfer functions in y or x directions. If not
        provided, they are assumed to be 2i * pi * u (or v), which are the
        effective transfer functions in Fourier differentiation.
    bc_center: float
        The value of the boundary condition at the center of the image.

    Returns
    -------
    Tensor
        The integrated image.
    """
    u, v = torch.fft.fftfreq(grad_x.shape[0]), torch.fft.fftfreq(grad_x.shape[1])
    u, v = torch.meshgrid(u, v, indexing="ij")
    if tf_y is None or tf_x is None:
        tf_y = 2j * torch.pi * u
        tf_x = 2j * torch.pi * v
    f_grad_y = pmath.fft2_precise(grad_y)
    f_grad_x = pmath.fft2_precise(grad_x)
    img = (f_grad_y * tf_y + f_grad_x * tf_x) / (tf_y.abs() ** 2 + tf_x.abs() ** 2 + 1e-5)
    img = -pmath.ifft2_precise(img)
    img = img + bc_center - img[img.shape[0] // 2, img.shape[1] // 2]
    return img


@timer()
def integrate_image_2d(grad_y: Tensor, grad_x: Tensor, bc_center: float = 0) -> Tensor:
    """
    Integrate an image with the gradient in y and x directions.

    Parameters
    ----------
    grad_y : Tensor
        The gradient in y direction.
    grad_x : Tensor
        The gradient in x direction.
    bc_center : float
        The boundary condition at the center of the image, by default 0

    Returns
    -------
    Tensor
        The integrated image.
    """
    left_boundary = torch.cumsum(grad_y[:, 0], dim=0)
    int_img = torch.cumsum(grad_x, dim=1) + left_boundary[:, None]
    int_img = int_img + bc_center - int_img[int_img.shape[0] // 2, int_img.shape[1] // 2]
    return int_img


@timer()
def convolve2d(
    image: Tensor,
    kernel: Tensor,
    padding: Literal["same", "valid"] = "same",
    padding_mode: Literal["replicate", "constant"] = "replicate",
) -> Tensor:
    """
    2D convolution with an explicitly given kernel using torch.nn.functional.conv2d.

    This routine flips the kernel to adhere with the textbook definition of convolution.
    torch.nn.functional.conv2d does not flip the kernel in itself.

    Parameters
    ----------
    image : Tensor
        A (... H, W) tensor of images. If the number of dimensions is greater than 2,
        the last two dimensions are interpreted as height and width, respectively.
    kernel : Tensor
        A (H, W) tensor of kernel.

    Returns
    -------
    Tensor
        A (... H, W) tensor of convolved images.
    """
    if not image.ndim >= 2:
        raise ValueError("Image must have at least 2 dimensions.")
    if not kernel.ndim == 2:
        raise ValueError("Kernel must have exactly 2 dimensions.")
    if not (kernel.shape[-2] % 2 == 1 and kernel.shape[-1] % 2 == 1):
        raise ValueError("Kernel dimensions must be odd.")

    if image.dtype.is_complex:
        kernel = kernel.type(image.dtype)

    # Reshape image to (N, 1, H, W).
    orig_shape = image.shape
    image = image.reshape(-1, 1, image.shape[-2], image.shape[-1])

    # Reshape kernel to (1, 1, H, W).
    kernel = kernel.flip((0, 1))
    kernel = kernel.reshape(1, 1, kernel.shape[-2], kernel.shape[-1])

    if padding == "same":
        pad_lengths = [
            kernel.shape[-1] // 2,
            kernel.shape[-1] // 2,
            kernel.shape[-2] // 2,
            kernel.shape[-2] // 2,
        ]
        image = torch.nn.functional.pad(image, pad_lengths, mode=padding_mode)

    result = torch.nn.functional.conv2d(image, kernel, padding="valid")
    result = result.reshape(*orig_shape[:-2], result.shape[-2], result.shape[-1])
    return result


@timer()
def convolve1d(
    input: Tensor,
    kernel: Tensor,
    padding: Literal["same", "valid"] = "same",
    padding_mode: Literal["replicate", "constant"] = "replicate",
    dim: int = -1,
) -> Tensor:
    """
    1D convolution with an explicitly given kernel using torch.nn.functional.conv1d.

    This routine flips the kernel to adhere with the textbook definition of convolution.
    torch.nn.functional.conv1d does not flip the kernel in itself.

    Parameters
    ----------
    image : Tensor
        A (... d) tensor of signals.
    kernel : Tensor
        A (d,) tensor of kernel.

    Returns
    -------
    Tensor
        A (... d) tensor of convolved signals.
    """
    if not input.ndim >= 1:
        raise ValueError("Image must have at least 1 dimensions.")
    if not kernel.ndim == 1:
        raise ValueError("Kernel must have exactly 1 dimensions.")

    if input.dtype.is_complex:
        kernel = kernel.type(input.dtype)

    dim = dim % input.ndim
    orig_shape = input.shape
    # Move dim to the end.
    if dim != input.ndim - 1:
        input = torch.moveaxis(input, dim, input.ndim - 1)
    bcast_shape = input.shape[:-1]
    # Reshape image to (N, 1, d).
    if input.ndim == 1:
        input = input.reshape(1, 1, input.shape[-1])
    else:
        input = input.reshape(-1, 1, input.shape[-1])

    # Reshape kernel to (1, 1, d).
    kernel = kernel.flip((0,))
    kernel = kernel.reshape(1, 1, kernel.shape[-1])

    if padding == "same":
        pad_lengths = [
            kernel.shape[-1] // 2,
            kernel.shape[-1] // 2,
        ]
        if kernel.shape[-1] % 2 == 0:
            pad_lengths[-1] -= 1
        input = torch.nn.functional.pad(input, pad_lengths, mode=padding_mode)

    result = torch.nn.functional.conv1d(input, kernel, padding="valid")

    # Restore shape.
    if len(orig_shape) == 1:
        result = result.reshape(orig_shape[0])
    else:
        result = result.reshape([*bcast_shape, result.shape[-1]])
        if dim != input.ndim - 1:
            result = torch.moveaxis(result, result.ndim - 1, dim)
    return result


@timer()
def gaussian_filter(image, sigma=1, size=3):
    x = torch.arange(-((size - 1) / 2), -((size - 1) / 2) + size, 1)
    gauss_1d = torch.exp(-0.5 * (x / sigma) ** 2)
    kernel = gauss_1d[:, None] * gauss_1d[None, :]
    kernel = kernel / torch.sum(kernel)
    return convolve2d(image, kernel, padding="same")


@timer()
def find_cross_corr_peak(
    f: Tensor,
    g: Tensor,
    scale: int,
    initial_guess=[0, 0],
    real_space_width: float = 3,
    dtype=torch.complex128,
):
    """
    Find the cross-correlation peak of two 2D arrays with arbitarily high precision.

    This implementation is based on:
    - Efficient subpixel image registration algorithms (2008) - Manuel Guizar-Sicairos

    The matrix multiplication inverse FT calculation was inferred from formula (12) in the paper:
    - Fast computation of Lyot-style coronagraph propagation (2007) - R. Soummer

    Note: position correction does not work properly when dtype is complex64 instead of complex128.
    The position correction will "walk off" if complex64 is used.

    Parameters
    ----------
    f : Tensor
        a (h, w) tensor.
    g : Tensor
        a (h, w) tensor.
    scale : int
        an integer specifying how much to upsample the cross-correlation.
    initial_guess : list[int]
        an initial guess of the location of the cross-correlation peak.
    real_space_width : float
        a number specifying the width of the cross-correlation (in units of non-upsampled pixels).

    Returns
    -------
    Tensor
        a tensor of the location of the cross-correlation peak.
    """
    f = torch.fft.ifftshift(f)
    g = torch.fft.ifftshift(g)

    dtype_real = torch.tensor(1 + 1j, dtype=dtype).real.dtype

    M, N = f.shape
    u = torch.fft.fftfreq(M, 1, dtype=dtype_real)[:, None]
    v = torch.fft.fftfreq(N, 1, dtype=dtype_real)[:, None]

    F = torch.fft.fft2(f.to(dtype))
    G = torch.fft.fft2(g.to(dtype))

    FG = F * G.conj()

    num_pts = int(real_space_width * scale / 2) * 2 + 1
    span = torch.linspace(-real_space_width / 2, real_space_width / 2, num_pts, dtype=dtype_real)

    x = initial_guess[0] + span[:, None]
    y = initial_guess[1] + span[:, None]

    # Do inverse FFT using matrix multiplication
    y_exp = torch.exp(2j * torch.pi * torch.matmul(v, y.transpose(1, 0))).to(dtype)
    tmp = torch.matmul(FG, y_exp)
    x_exp = torch.exp(2j * torch.pi * torch.matmul(x, u.transpose(1, 0))).to(dtype)
    corr = torch.matmul(x_exp, tmp)

    max_idx = torch.tensor(torch.unravel_index(torch.argmax(corr.abs()), corr.shape))
    est_shift = torch.tensor([x[max_idx[0]], y[max_idx[1]]])

    return est_shift


@timer()
def total_variation_2d_chambolle(
    image: Tensor, lmbda: float = 0.01, niter: int = 2, tau: float = 0.125
) -> Tensor:
    """
    Apply total variation constraint to an image using Chambolle's total variation projection method
    (https://doi.org/10.5201/ipol.2013.61).

    Parameters
    ----------
    image : Tensor
        A (H, W) tensor.
    lmbda : float
        The trade-off parameter (1 / lambda in the paper).
        Larger lmbda means stronger smoothing.
    niter : int
        An integer specifying the number of iterations.
    tau : float
        The time-step parameter.

    Returns
    -------
    Tensor
        A (H, W) tensor.
    """

    def div(p):
        py = p[..., 0]
        px = p[..., 1]
        fy = nearest_neighbor_gradient(py, "backward", dim=0)[0]
        fx = nearest_neighbor_gradient(px, "backward", dim=1)[1]
        fd = fx + fy
        return fd

    if image.ndim != 2:
        raise ValueError("Input tensor must be two dimensional")

    x0 = image
    xi = torch.zeros(list(image.shape) + [2], dtype=image.dtype, device=image.device)

    for _ in range(niter):
        # Chambolle step
        grad_y, grad_x = nearest_neighbor_gradient(div(xi) - image / lmbda, "forward")
        gdv = torch.stack([grad_y, grad_x], dim=-1)

        # Isotropic
        d = torch.sqrt(torch.sum(gdv**2, dim=-1, keepdim=True))

        xi = (xi + tau * gdv) / (1 + tau * d)

        # Reconstruct
        image = image - lmbda * div(xi)

    # Prevent pushing values to 0
    image = torch.sum(x0 * image) / torch.sum(image**2) * image
    return image


@timer()
def remove_grid_artifacts(
    img: Tensor,
    pixel_size_m: float,
    period_x_m: float,
    period_y_m: float,
    window_size: int,
    direction: Literal["x", "y", "xy"] = "xy",
):
    """
    Remove grid artifacts by setting the region near each harmonic peak
    corresponding to the set periodicity of the artifacts to 0 in the input
    image's Fourier domain.

    Adapted from fold_slice (remove_grid_artifact.m).

    Parameters
    ----------
    img : Tensor
        The input image.
    pixel_size_m : float
        The pixel size in meter.
    step_size_x : float
        The period of the artifacts in the x direction in meter.
    step_size_y : float
        The period of the artifacts in the y direction in meter.
    window_size : int
        The radius of the region around each harmonic peak in which the values
        are to be set to 0, given in pixels.
    direction : {'x', 'y', 'xy'}
        The direction of the artifacts to be removed.

    Returns
    -------
    Tensor
        The output image.
    """
    if img.ndim != 2:
        raise ValueError("Input tensor must be two dimensional.")

    ny, nx = img.shape
    dk_y, dk_x = 1 / pixel_size_m / ny, 1 / pixel_size_m / nx
    center_y, center_x = math.floor(ny / 2) + 1, math.floor(nx / 2) + 1

    k_max = 0.5 / pixel_size_m
    f_img = torch.fft.fftshift(pmath.fft2_precise(img))
    # Frequencies of the artifacts.
    dk_s_y, dk_s_x = 1 / period_y_m, 1 / period_x_m

    x_range, y_range = 0, 0
    # Get the frequencies of all harmonic peaks.
    if "x" in direction:
        x_range = torch.arange(math.ceil(-k_max / dk_s_x), math.floor(k_max / dk_s_x))
    if "y" in direction:
        y_range = torch.arange(math.ceil(-k_max / dk_s_y), math.floor(k_max / dk_s_y))

    for i in range(len(y_range)):
        for j in range(len(x_range)):
            # Avoid DC.
            if not (x_range[j] == 0 and y_range[i] == 0):
                window_y_lb = int(
                    max(torch.round(y_range[i] * dk_s_y / dk_y) + center_y - window_size, 0)
                )
                window_y_ub = int(
                    min(torch.round(y_range[i] * dk_s_y / dk_y) + center_y + window_size, ny)
                )

                window_x_lb = int(
                    max(torch.round(x_range[j] * dk_s_x / dk_x) + center_x - window_size, 0)
                )
                window_x_ub = int(
                    min(torch.round(x_range[j] * dk_s_x / dk_x) + center_x + window_size, nx)
                )

                f_img[window_y_lb:window_y_ub, window_x_lb:window_x_ub] = 0

    img_new = torch.real(torch.fft.ifft2(torch.fft.ifftshift(f_img)))
    return img_new


@timer()
def median_filter_1d(x: Tensor, window_size: int = 5):
    """
    Apply a median filter to a 1D array.

    Parameters
    ----------
    x : Tensor
        A (..., N) tensor.
    window_size : int
        The size of the window.

    Returns
    -------
    Tensor
        The filtered array.
    """
    y = x.detach().clone()
    rad = window_size // 2
    for i in range(rad, x.shape[-1] - rad):
        y[..., i] = torch.median(x[..., i - rad : i - rad + window_size], dim=-1).values
    return y


@timer()
def generate_vignette_mask(
    shape: tuple[int, int], 
    margin: int = 20, 
    sigma: float = 1.0, 
    method: Literal["gaussian", "linear"] = "gaussian"
):
    """
    Generate a vignette mask for an image of shape `shape`.
    """
    mask = torch.ones(shape, device=torch.get_default_device())
    mask = vignette(mask, margin, sigma, method=method)
    return mask


@timer()
def vignette(
    img: Tensor, 
    margin: int = 20, 
    sigma: float = 1.0, 
    method: Literal["gaussian", "linear", "window"] = "gaussian",
    dim=(-2, -1)):
    """
    Vignett an image so that it gradually decays near the boundary.
    For each dimension of the image, a mask with a width of `2 * margin`
    and with half of it filled with 0s and half with 1s is
    generated and convolved with a Gaussian kernel of size
    `margin` and standard deviation `sigma`. The blurred mask is cropped and
    multiplied to the near-edge regions of the image.

    This function is not differentiable because of the clone and
    slice-assignment operation.

    Parameters
    ----------
    img : Tensor
        The input image.
    margin : int
        The margin of image where the decay takes place. 
        Only used if `method` is "gaussian" or "linear".
    sigma : float
        The standard deviation of the Gaussian kernel. 
        Only used if `method` is "gaussian".
    method : Literal["gaussian", "linear", "window"]
        The method to use to generate the vignette mask. 
        "window" is a Hann window.
    """
    dims = [d % img.ndim for d in dim]
    img = img.clone()
    for i_dim in dims:
        if img.shape[i_dim] <= 2 * margin:
            continue
        mask_shape = (
            [img.shape[i] for i in range(i_dim)]
            + [2 * margin]
            + [img.shape[i] for i in range(i_dim + 1, img.ndim)]
        )
        if method == "gaussian":
            mask = torch.zeros(mask_shape, device=img.device)
            mask_slicer = [slice(None)] * i_dim + [slice(margin, None)]
            mask[*mask_slicer] = 1.0
            gauss_win = torch.signal.windows.gaussian(margin // 2, std=sigma)
            gauss_win = gauss_win / torch.sum(gauss_win)
            mask = convolve1d(mask, gauss_win, dim=i_dim, padding="same")
            mask_final_slicer = [slice(None)] * i_dim + [slice(len(gauss_win), len(gauss_win) + margin)]
            mask = mask[*mask_final_slicer]
            mask = torch.where(mask < 1e-3, 0, mask)
        elif method == "linear":
            ramp = torch.linspace(0, 1, margin)
            new_shape = [1] * img.ndim
            new_shape[i_dim] = margin
            ramp = ramp.reshape(new_shape)
            rep = list(img.shape)
            rep[i_dim] = 1
            mask = ramp.repeat(rep)
        elif method == "window":
            window_func = torch.hamming_window(
                img.shape[i_dim], periodic=True, alpha=0.5, beta=0.5
            )
            img = img * window_func.reshape([-1] + [1] * (img.ndim - i_dim - 1))
        else:
            raise ValueError(f"Unknown method: {method}")

        if method in ["gaussian", "linear"]:
            slicer = [slice(None)] * i_dim + [slice(0, margin)]
            img[slicer] = img[slicer] * mask

            slicer = [slice(None)] * i_dim + [slice(-margin, None)]
            img[slicer] = img[slicer] * mask.flip(i_dim)
    return img


@timer()
def remove_polynomial_background(
    img: Tensor,
    flat_region_mask: Tensor,
    polyfit_order: int = 1,
) -> Tensor:
    """
    Fit a 2D polynomial to the region that is supposed to be flat in an
    image, and subtract the fitted function from the image.

    Parameters
    ----------
    img : Tensor
        The input image.
    flat_region_mask : Tensor
        A boolean mask with the same shape as `img` that specifies the region of
        the image that should be flat.
    polyfit_order : int, optional
        The order of the polynomial to fit. Should be an integer >= 0. If 0,
        just subtract the average. The default is 1.

    Returns
    -------
    Tensor
        The image with the polynomial background subtracted.
    """
    if polyfit_order == 0:
        return img - img[flat_region_mask].mean()
    ys, xs = torch.where(flat_region_mask)
    y_full, x_full = torch.meshgrid(
        torch.arange(img.shape[0]), torch.arange(img.shape[1]), indexing="ij"
    )
    y_full = y_full.reshape(-1)
    x_full = x_full.reshape(-1)

    y_all_orders = []
    x_all_orders = []
    y_full_all_orders = []
    x_full_all_orders = []
    for order in range(polyfit_order + 1):
        y_all_orders.append(ys**order)
        x_all_orders.append(xs**order)
        y_full_all_orders.append(y_full**order)
        x_full_all_orders.append(x_full**order)
    const_basis = torch.ones(len(ys), device=img.device)
    const_basis_full = torch.ones(len(y_full), device=img.device)

    a_mat = torch.stack(y_all_orders + x_all_orders + [const_basis], dim=1)
    b_vec = img[flat_region_mask].reshape(-1, 1)
    x_vec = torch.linalg.solve(a_mat, b_vec)
    a_mat_full = torch.stack(y_full_all_orders + x_full_all_orders + [const_basis_full], dim=1)
    bg = a_mat_full @ x_vec
    bg = bg.reshape(img.shape)
    return img - bg


@timer()
def unwrap_phase_2d(
    img: Tensor,
    fourier_shift_step: float = 0.5,
    image_grad_method: Literal[
        "fourier_shift", "fourier_differentiation", "nearest"
    ] = "fourier_differentiation",
    image_integration_method: Literal[
        "fourier", "discrete", "deconvolution"
    ] = "fourier",
    weight_map: Optional[Tensor] = None,
    flat_region_mask: Optional[Tensor] = None,
    deramp_polyfit_order: int = 1,
    return_phase_grads: bool = False,
    eps: float = 1e-9,
):
    """
    Unwrap phase of 2D image.

    Parameters
    ---------- 
    img : Tensor
        A complex 2D tensor giving the image.
    fourier_shift_step : float
        The finite-difference step size used to calculate the gradient,
        if the Fourier shift method is used.
    image_grad_method : str
        The method used to calculate the phase gradient.
            - "fourier_shift": Use Fourier shift to perform shift.
            - "nearest": Use nearest neighbor to perform shift.
            - "fourier_differentiation": Use Fourier differentiation.
    image_integration_method : str
        The method used to integrate the image back from gradients.
            - "fourier": Use Fourier integration as implemented in PtychoShelves.
            - "deconvolution": Deconvolve ramp filter.
            - "discrete": Use cumulative sum.
    weight_map : Optional[Tensor]
        A weight map multiplied to the input image.
    flat_region_mask : Optional[Tensor]
        A boolean mask with the same shape as `img` that specifies the region of
        the image that should be flat. This is used to remove unrealistic phase
        ramps in the image. If None, de-ramping will not be done.
    deramp_polyfit_order : int
        The order of the polynomal fit used to de-ramp the phase.
    return_phase_grads : bool
        Whether to return the phase gradient.
    eps : float
        A small number to avoid division by zero.

    Returns
    -------
    Tensor
        The phase of the original image after unwrapping.
    """
    if not img.is_complex():
        raise ValueError("Input tensor must be complex.")

    if isinstance(weight_map, Tensor):
        weight_map = torch.clip(weight_map, 0.0, 1.0)
    else:
        weight_map = 1

    img = weight_map * img / (img.abs() + eps)
    bc_center = torch.angle(img[img.shape[0] // 2, img.shape[1] // 2])

    # Pad image to avoid FFT boundary artifacts.
    padding = [64, 64]
    if torch.any(torch.tensor(padding) > 0):
        if img.shape[-2] > padding[0] and img.shape[-1] > padding[1]:
            padding_mode = "reflect"
        else:
            padding_mode = "replicate"
        img = torch.nn.functional.pad(
            img[None, None, :, :], (padding[1], padding[1], padding[0], padding[0]), mode=padding_mode
        )[0, 0]
        img = vignette(img, margin=10, sigma=2.5)

    gy, gx = get_phase_gradient(
        img, fourier_shift_step=fourier_shift_step, image_grad_method=image_grad_method
    )

    if image_integration_method == "discrete" and torch.any(torch.tensor(padding) > 0):
        gy = gy[padding[0] : -padding[0], padding[1] : -padding[1]]
        gx = gx[padding[0] : -padding[0], padding[1] : -padding[1]]
    if image_integration_method == "discrete":
        phase = torch.real(integrate_image_2d(gy, gx, bc_center=bc_center))
    elif image_integration_method == "fourier":
        phase = torch.real(integrate_image_2d_fourier(gy, gx))
    elif image_integration_method == "deconvolution":
        phase = torch.real(integrate_image_2d_deconvolution(gy, gx, bc_center=bc_center))
    else:
        raise ValueError(f"Unknown integration method: {image_integration_method}")
    
    if image_integration_method != "discrete" and torch.any(torch.tensor(padding) > 0):
        gy = gy[padding[0] : -padding[0], padding[1] : -padding[1]]
        gx = gx[padding[0] : -padding[0], padding[1] : -padding[1]]
        phase = phase[padding[0] : -padding[0], padding[1] : -padding[1]]

    if flat_region_mask is not None:
        phase = remove_polynomial_background(
            phase, flat_region_mask, polyfit_order=deramp_polyfit_order
        )
    if return_phase_grads:
        return phase, (gy, gx)
    return phase


@timer()
def find_center_of_mass(img: Tensor):
    """
    Find the center of mass of one or a stack of 2D images.
    
    Parameters
    ----------
    img : Tensor
        A (N, H, W) or (H, W) tensor giving a batch of images or a single image.
    
    Returns
    -------
    Tensor
        A (N, 2) or (2,) tensor giving the centers of mass.
    """
    orig_ndim = img.ndim
    if img.ndim == 2:
        img = img[None]
    
    if img.is_complex():
        img = torch.abs(img)
    
    # Find the centers of mass of the last 2 dimensions.
    y, x = torch.meshgrid(torch.arange(img.shape[-2]), torch.arange(img.shape[-1]), indexing="ij")
    y = y.to(img.device)
    x = x.to(img.device)
    r = torch.stack((y, x), dim=-1)
    com = (r[None] * img[..., None]).sum(dim=(-3, -2)) / img.sum(dim=(-2, -1))[..., None]
    if orig_ndim == 2:
        return com[0]
    return com


@timer()
def central_crop(img: Tensor, crop_size: tuple[int, int]) -> Tensor:
    """
    Crop the center of an image.
    
    Parameters
    ----------
    img : Tensor
        A (..., H, W) tensor giving the input image(s).
    crop_size : tuple[int, int]
        crop size.
    
    Returns
    -------
    Tensor
        The image cropped to the target size.
    """
    return img[
        ..., 
        img.shape[-2] // 2 - crop_size[0] // 2 : img.shape[-2] // 2 - crop_size[0] // 2 + crop_size[0], 
        img.shape[-1] // 2 - crop_size[1] // 2 : img.shape[-1] // 2 - crop_size[1] // 2 + crop_size[1]
    ]


@timer()
def central_pad(
    img: Tensor, 
    target_size: tuple[int, int], 
    mode: Literal["constant", "reflect", "replicate", "circular"] = "constant", 
    value: float = 0.0
) -> Tensor:
    """
    Pad the center of an image.
    
    Parameters
    ----------
    img : Tensor
        A (..., H, W) tensor giving the input image(s).
    target_size : tuple[int, int]
        target size.
    
    Returns
    -------
    Tensor
        The image padded to the target size.
    """
    pad_size = [
        (target_size[0] - img.shape[-2]) // 2, target_size[-2] - (target_size[0] - img.shape[-2]) // 2 - img.shape[-2],
        (target_size[1] - img.shape[-1]) // 2, target_size[-1] - (target_size[1] - img.shape[-1]) // 2 - img.shape[-1]
    ]
    return torch.nn.functional.pad(img, (pad_size[2], pad_size[3], pad_size[0], pad_size[1]), mode=mode, value=value)


@timer()
def central_crop_or_pad(img: Tensor, target_size: tuple[int, int]) -> Tensor:
    """
    Crop or pad the center of an image to the target size.
    
    Parameters
    ----------
    img : Tensor
        A (..., H, W) tensor giving the input image(s).
    target_size : tuple[int, int]
        target size.
    
    Returns
    -------
    Tensor
        The image cropped or padded to the target size.
    """
    for i in range(2):
        target_size_current_dim = [img.shape[-2]] * (i == 1) + [target_size[i]] + [img.shape[-1]] * (i == 0)
        if img.shape[-2 + i] > target_size[-2 + i]:
            img = central_crop(img, target_size_current_dim)
        elif img.shape[-2 + i] < target_size[-2 + i]:
            img = central_pad(img, target_size_current_dim)
    return img



# Copyright © 2025 UChicago Argonne, LLC All right reserved
# Full license accessible at https://github.com//AdvancedPhotonSource/pty-chi/blob/main/LICENSE

from typing import Optional
import dataclasses
from dataclasses import field
import logging

import ptychi.api.options.base as base
import ptychi.api.options.task as task_options
import ptychi.api.enums as enums

logger = logging.getLogger(__name__)


@dataclasses.dataclass
class LSQMLReconstructorOptions(base.ReconstructorOptions):
    
    noise_model: enums.NoiseModels = enums.NoiseModels.GAUSSIAN
    """
    The noise model to use.
    """
    
    gaussian_noise_std: float = 0.5
    """
    The standard deviation of the gaussian noise. Only used when `noise_model == enums.NoiseModels.GAUSSIAN`.
    """
    
    solve_obj_prb_step_size_jointly_for_first_slice_in_multislice: bool = False
    """
    Whether to solve the simultaneous object/probe step length calculation;
    in FoldSlice they use independent (non-joint) step length calculation, but 
    we're adding the option of using simultaneous AND non-simultaneous step 
    length calculation.
    """
    
    solve_step_sizes_only_using_first_probe_mode: bool = True
    """
    If True, object and probe step sizes will only be calculated using the first probe mode.
    This is how it is done in PtychoShelves.
    """
    
    momentum_acceleration_gain: float = 0.0
    """The gain of momentum acceleration. If 0, momentum acceleration is not used."""
    
    momentum_acceleration_gradient_mixing_factor: Optional[float] = 1
    """
    Controls how the current gradient is mixed with the accumulated velocity in LSQML
    momentum acceleration:
    
    `velocity = (1 - friction) * velocity + momentum_acceleration_gradient_mixing_factor * delta_o`
    
    If None, this mixing factor is automatically chosen to be `friction`:
    
    `velocity = (1 - friction) * velocity + friction * delta_o`
    
    Using `None` usually provides better stability. However, it may cause the speed of convergence to be
    too slow in some cases. Set this parameter to 1 to reproduce the behavior in PtychoShelves.
    """
    
    rescale_probe_intensity_in_first_epoch: bool = True
    """
    If True, probe intensity is rescaled in the first epoch using the average intensity of all
    diffraction patterns. Set this to False if you want the probe intensity to stay constant.
    You may also want to check `ObjectOptions.remove_object_probe_ambiguity`.
    """
    
    preconditioning_damping_factor: float = 0.1
    """
    The damping factor for applying preconditioning to the object update, which is calculated as::
    
        delta_o_hat = delta_o_hat / torch.sqrt(preconditioner ** 2 + (preconditioner.max() * mixing_factor) ** 2)
    """
    
    def check(self, options: "LSQMLOptions"):
        super().check(options)
        if self.rescale_probe_intensity_in_first_epoch:
            if options.probe_options.power_constraint.enabled:
                logger.warning(
                    "`rescale_probe_intensity_in_first_epoch` and `ProbeOptions.power_constraint` "
                    "are both enabled, which may lead to unexpected results."
                )
    
    def get_reconstructor_type(self) -> enums.Reconstructors:
        return enums.Reconstructors.LSQML
    

@dataclasses.dataclass
class LSQMLObjectOptions(base.ObjectOptions):
    
    optimal_step_size_scaler: float = 0.9
    """
    A scaler for the solved optimal step size (beta_LSQ in PtychoShelves).
    """
    
    multimodal_update: bool = True
    """
    If True, object update direction is calculated and summed over all probe modes. 
    Otherwise, only the first mode will be used for object update. However, forward
    propagation always uses all probe modes regardless of this option.
    """


@dataclasses.dataclass
class LSQMLProbeOptions(base.ProbeOptions):
    optimal_step_size_scaler: float = 0.9
    """
    A scaler for the solved optimal step size (beta_LSQ in PtychoShelves).
    """
    
    
@dataclasses.dataclass
class LSQMLProbePositionOptions(base.ProbePositionOptions):
    pass


@dataclasses.dataclass
class LSQMLOPRModeWeightsOptions(base.OPRModeWeightsOptions):
    pass


@dataclasses.dataclass
class LSQMLOptions(task_options.PtychographyTaskOptions):

    reconstructor_options: LSQMLReconstructorOptions = field(default_factory=LSQMLReconstructorOptions)
    
    object_options: LSQMLObjectOptions = field(default_factory=LSQMLObjectOptions)
    
    probe_options: LSQMLProbeOptions = field(default_factory=LSQMLProbeOptions)
    
    probe_position_options: LSQMLProbePositionOptions = field(default_factory=LSQMLProbePositionOptions)
    
    opr_mode_weight_options: LSQMLOPRModeWeightsOptions = field(default_factory=LSQMLOPRModeWeightsOptions)



# Copyright © 2025 UChicago Argonne, LLC All right reserved
# Full license accessible at https://github.com//AdvancedPhotonSource/pty-chi/blob/main/LICENSE

from enum import StrEnum, auto


class BatchingModes(StrEnum):
    RANDOM = auto()
    COMPACT = auto()
    UNIFORM = auto()


class Optimizers(StrEnum):
    SGD = auto()
    ADAM = auto()
    RMSPROP = auto()
    ADAGRAD = auto()
    ADADELTA = auto()
    LBFGS = auto()
    ASGD = auto()
    SPARSE_ADAM = auto()
    ADAMAX = auto()
    RADAM = auto()
    ADAMW = auto()


class LossFunctions(StrEnum):
    MSE = auto()
    POISSON = auto()
    MSE_SQRT = auto()


class Reconstructors(StrEnum):
    base = auto()
    AD_GENERAL = auto()
    AD_PTYCHO = auto()
    LSQML = auto()
    PIE = auto()
    EPIE = auto()
    RPIE = auto()
    DM = auto()
    BH = auto()


class PositionCorrectionTypes(StrEnum):
    GRADIENT = auto()
    CROSS_CORRELATION = auto()


class NoiseModels(StrEnum):
    GAUSSIAN = auto()
    POISSON = auto()


class OrthogonalizationMethods(StrEnum):
    GS = auto()
    SVD = auto()


class ForwardModels(StrEnum):
    base = auto()
    PLANAR_PTYCHOGRAPHY = auto()


class Directions(StrEnum):
    X = auto()
    Y = auto()
    XY = auto()


class Devices(StrEnum):
    CPU = auto()
    GPU = auto()


class Dtypes(StrEnum):
    FLOAT32 = auto()
    FLOAT64 = auto()


class ImageGradientMethods(StrEnum):
    FOURIER_DIFFERENTIATION = auto()
    FOURIER_SHIFT = auto()
    NEAREST = auto()
    GAUSSIAN = auto()
    
class ImageIntegrationMethods(StrEnum):
    FOURIER = auto()
    DECONVOLUTION = auto()
    DISCRETE = auto()


class PatchInterpolationMethods(StrEnum):
    FOURIER = auto()
    BILINEAR = auto()
    NEAREST = auto()


class OPRWeightSmoothingMethods(StrEnum):
    MEDIAN = auto()
    POLYNOMIAL = auto()


class AffineDegreesOfFreedom(StrEnum):
    TRANSLATION = auto()
    ROTATION = auto()
    SCALE = auto()
    SHEAR = auto()
    ASYMMETRY = auto()


class DIPModels(StrEnum):
    UNET = auto()
    AUTOENCODER = auto()


class ObjectPosOriginCoordsMethods(StrEnum):
    POSITIONS = auto()
    SUPPORT = auto()
    SPECIFIED = auto()



# Copyright © 2025 UChicago Argonne, LLC All right reserved
# Full license accessible at https://github.com//AdvancedPhotonSource/pty-chi/blob/main/LICENSE

import dataclasses
from dataclasses import field

import ptychi.api.options.base as base
import ptychi.api.options.task as task_options
import ptychi.api.enums as enums


@dataclasses.dataclass
class DMReconstructorOptions(base.ReconstructorOptions):
    def get_reconstructor_type(self) -> enums.Reconstructors:
        return enums.Reconstructors.DM

    exit_wave_update_relaxation: float = 1
    "Relaxation multiplier for the exit wave update."

    chunk_length: int = 1
    """Number of scan points used in each chunk of the difference map exit wave update loop.
    Smaller values are more memory efficient, but can be slower."""


@dataclasses.dataclass
class DMObjectOptions(base.ObjectOptions):
    amplitude_clamp_limit: float = 1000
    """Maximum allowed amplitude for the object reconstruction. Values above this will be clamped 
    to this value."""

    inertia: float = 0
    "Inertia of the object update. Should be between 0 and 1."


@dataclasses.dataclass
class DMProbeOptions(base.ProbeOptions):
    inertia: float = 0
    "Inertia of the probe update. Should be between 0 and 1."


@dataclasses.dataclass
class DMProbePositionOptions(base.ProbePositionOptions):
    pass


@dataclasses.dataclass
class DMOPRModeWeightsOptions(base.OPRModeWeightsOptions):
    pass


@dataclasses.dataclass
class DMOptions(task_options.PtychographyTaskOptions):
    reconstructor_options: DMReconstructorOptions = field(default_factory=DMReconstructorOptions)

    object_options: DMObjectOptions = field(default_factory=DMObjectOptions)

    probe_options: DMProbeOptions = field(default_factory=DMProbeOptions)

    probe_position_options: DMProbePositionOptions = field(default_factory=DMProbePositionOptions)

    opr_mode_weight_options: DMOPRModeWeightsOptions = field(
        default_factory=DMOPRModeWeightsOptions
    )



# Copyright © 2025 UChicago Argonne, LLC All right reserved
# Full license accessible at https://github.com//AdvancedPhotonSource/pty-chi/blob/main/LICENSE

import time
from functools import wraps
from typing import Callable, TypeVar, Optional, List, Dict, Union
import numpy as np
import matplotlib.pyplot as plt
import copy
import torch
from collections import defaultdict


# Type variables to retain function signatures
T = TypeVar("T", bound=Callable)

ENABLE_TIMING = False
"Global flag to enable or disable timing."
ELAPSED_TIME_DICT: dict[str, np.ndarray] = defaultdict(lambda: np.array([]))
"""
A dictionary containing numpy arrays of the measured execution times of 
each timed function.
"""
ADVANCED_TIME_DICT: dict[str, Union[np.ndarray, dict]] = defaultdict(lambda: {})
"""
A nested dictionary, where each level of the dictionary contains
1) a key, value pair ("time": np.ndarray) that contains all measured
execution times for that function and 2) zero or more key-value 
pairs (function_name: dict) where function_name refers to the name
of each functions called in the function currently being times.

Note that only functions with the `timer` decorator will show up
in `ADVANCED_TIME_DICT`.
"""
CURRENT_DICT_REFERENCE = ADVANCED_TIME_DICT  # Initialized to the top level of ADVANCED_TIME_DICT
"""
A reference to the level of `ADVANCED_TIME_DICT` that corresponds to 
the function currently being executed.
"""
TIMING_OVERHEAD_ARRAY = np.array([])
"""
A numpy array containing measurements of how long it takes to execute 
functions in the `timer` decorator.
"""
default_label_font_size = 8


def toggle_timer(enable: bool):
    """
    Toggle the global ENABLE_TIMING flag.

    Parameters
    ----------
    enable : bool
        If True, enable timing. If False, disable timing.
    """
    global ENABLE_TIMING
    ENABLE_TIMING = enable


def timer(enabled: bool = True, override_with_name: Optional[str] = None):
    """
    Decorator to time a function's execution time and the execution time of the timed code
    within that function. This function is enabled or disabled depending on the state of the
    global ENABLE_TIMING flag.

    The results of the timer function will be recorded in `ELAPSED_TIME_DICT` and
    `ADVANCED_TIME_DICT`.

    Parameters
    ----------
    enabled : bool, optional
        Whether timing is enabled for the decorated function. Default is True.
    override_with_name : str, optional
        Custom name to use for the function in the timing dictionary. If not
        specified, the function name is automatically generated.

    Returns
    -------
    Callable
        The wrapped function.
    """

    def decorator(func: T) -> T:
        @wraps(func)
        def wrapper(*args, **kwargs):
            if enabled and globals().get("ENABLE_TIMING", False):
                # Measure the overhead from running the timer function
                measure_overhead_start_1 = time.time()
                if override_with_name is None:
                    function_name = func.__qualname__
                else:
                    function_name = override_with_name
                saved_dict_reference = update_current_dict_reference(function_name)
                overhead_time_1 = time.time() - measure_overhead_start_1

                # Measure function execution time
                torch.cuda.synchronize()
                start_time = time.time()
                result = func(*args, **kwargs)
                torch.cuda.synchronize()
                elapsed_time = time.time() - start_time

                # Measure the overhead from running the timer function
                measure_overhead_start_2 = time.time()
                update_elapsed_time_dict(function_name, elapsed_time)
                update_advanced_time_dict(elapsed_time)
                # Traverse back up the advanced timing dicts
                revert_current_dict_reference(saved_dict_reference)
                global TIMING_OVERHEAD_ARRAY
                overhead_time_2 = time.time() - measure_overhead_start_2
                TIMING_OVERHEAD_ARRAY = np.append(
                    TIMING_OVERHEAD_ARRAY,
                    overhead_time_1 + overhead_time_2,
                )
            else:
                # If timing is disabled, just call the function
                result = func(*args, **kwargs)
            return result

        # Ensure the wrapper function has the same type as the original
        return wrapper  # type: ignore

    return decorator


class InlineTimer:
    """
    A timer class for inline timing of code blocks.

    Parameters
    ----------
    name : str
        The name associated with the timer that will be recorded
        in the timing dictionaries.
    enabled : bool, optional
        Whether the timer is enabled, by default True.
    """

    def __init__(self, name: str, enabled: bool = True):
        self.name = name
        self.enabled = enabled
        self.overhead_time = 0

    def start(self):
        """
        Starts the timer if timing is enabled.
        """
        if self.enabled and globals().get("ENABLE_TIMING", False):
            measure_overhead_start = time.time()
            saved_dict_reference = update_current_dict_reference(self.name)
            self.saved_dict_reference = saved_dict_reference
            self.overhead_time = time.time() - measure_overhead_start
            torch.cuda.synchronize()
            self.start_time = time.time()

    def end(self):
        """
        Stops the timer and records the elapsed time if timing is enabled.
        """
        if self.enabled and globals().get("ENABLE_TIMING", False):
            torch.cuda.synchronize()
            elapsed_time = time.time() - self.start_time
            measure_overhead_start = time.time()
            update_elapsed_time_dict(self.name, elapsed_time)
            update_advanced_time_dict(elapsed_time)
            revert_current_dict_reference(self.saved_dict_reference)
            global TIMING_OVERHEAD_ARRAY
            self.overhead_time += time.time() - measure_overhead_start
            TIMING_OVERHEAD_ARRAY = np.append(TIMING_OVERHEAD_ARRAY, self.overhead_time)


def update_elapsed_time_dict(function_name: str, elapsed_time: float):
    """
    Updates the global elapsed time dictionary with the elapsed time for a function.

    Parameters
    ----------
    function_name : str
        The name of the function being timed.
    elapsed_time : float
        The elapsed time for the function execution.
    """
    ELAPSED_TIME_DICT[function_name] = np.append(ELAPSED_TIME_DICT[function_name], elapsed_time)


def update_current_dict_reference(function_name: str) -> dict:
    """
    Updates the current reference in the advanced timing dictionary to a nested level.

    Parameters
    ----------
    function_name : str
        The name of the function being timed.

    Returns
    -------
    dict
        The previous dictionary reference.
    """
    global CURRENT_DICT_REFERENCE
    # Save the parent to traverse back to later
    saved_dict_reference = CURRENT_DICT_REFERENCE
    # Create new dict if necessary
    if function_name not in CURRENT_DICT_REFERENCE.keys():
        CURRENT_DICT_REFERENCE[function_name] = defaultdict(lambda: {})
        CURRENT_DICT_REFERENCE[function_name]["time"] = np.array([])
    # Update the pointer to the current dict
    CURRENT_DICT_REFERENCE = CURRENT_DICT_REFERENCE[function_name]
    return saved_dict_reference


def update_advanced_time_dict(elapsed_time: float):
    """
    Updates the advanced timing dictionary with the elapsed time.

    Parameters
    ----------
    elapsed_time : float
        The elapsed time for the function execution.
    """
    global CURRENT_DICT_REFERENCE
    CURRENT_DICT_REFERENCE["time"] = np.append(CURRENT_DICT_REFERENCE["time"], elapsed_time)


def revert_current_dict_reference(saved_dict_reference: dict):
    """
    Reverts the current dictionary reference in the advanced timing dictionary.

    Parameters
    ----------
    saved_dict_reference : dict
        The saved dictionary reference to revert to.
    """
    global CURRENT_DICT_REFERENCE
    CURRENT_DICT_REFERENCE = saved_dict_reference


def clear_timer_globals():
    """
    Clears the global timing dictionaries and resets the state.
    """
    global ELAPSED_TIME_DICT
    global ADVANCED_TIME_DICT
    global CURRENT_DICT_REFERENCE
    global TIMING_OVERHEAD_ARRAY
    ELAPSED_TIME_DICT = defaultdict(lambda: np.array([]))
    ADVANCED_TIME_DICT = defaultdict(lambda: {})
    CURRENT_DICT_REFERENCE = ADVANCED_TIME_DICT
    TIMING_OVERHEAD_ARRAY = np.array([])


def plot_elapsed_time_bar_plot(
    include: Optional[List[str]] = None,
    exclude: Optional[List[str]] = None,
    top_n: Optional[int] = None,
    figsize: Optional[tuple] = None,
    only_include_leafs: bool = False,
    advanced_time_dict: Optional[dict] = None,
    elapsed_time_dict: Optional[dict] = None,
    sort: bool = False,
):
    """
    Plots a simple bar chart of elapsed times for timed functions.

    Parameters
    ----------
    include : list of str, optional
        Specific functions to include in the plot, by default None.
    exclude : list of str, optional
        Specific functions to exclude from the plot, by default None.
    top_n : int, optional
        The top N functions with the largest elapsed times to include, by default None.
    only_include_leafs : bool, optional
        If True, includes only leaf functions in the plot, by default False.
    figsize : tuple, optional
        The figure size for the plot, by default None.
    label_fontsize: int, optional
        The font size for bar labels.
    advanced_time_dict : dict, optional
        By default, this will be the global `ADVANCED_TIME_DICT`. Override this
        to plot some other dict. The `ADVANCED_TIME_DICT` is only used when
        `only_include_leafs` is True.
    elapsed_time_dict : dict, optional
        By default, this will be the global `ELAPSED_TIME_DICT`. Override this
        to plot some other dict.
    """
    # If elapsed_time_dict was not passed in, use the
    # value from the global
    if elapsed_time_dict is None:
        global ELAPSED_TIME_DICT
        elapsed_time_dict = ELAPSED_TIME_DICT

    if only_include_leafs:
        if advanced_time_dict is None:
            global ADVANCED_TIME_DICT
            advanced_time_dict = ADVANCED_TIME_DICT
        elapsed_time_dict = select_leaf_functions(advanced_time_dict)

    elapsed_time_dict = return_dict_subset_copy(elapsed_time_dict, include, exclude)

    # Remove the prefix from keys and calculate sums
    sums = [np.sum(elapsed_time_dict[key]) for key in elapsed_time_dict.keys()]

    generate_time_barplot(
        times=sums,
        labels=list(elapsed_time_dict.keys()),
        colors="skyblue",
        title="Total elapsed time for each function",
        figsize=figsize,
        sort=sort,
        top_n=top_n,
    )


def plot_elapsed_time_bar_plot_advanced(
    function_name: str,
    include: Optional[List[str]] = None,
    exclude: Optional[List[str]] = None,
    top_n: Optional[int] = None,
    max_levels: int = None,
    use_long_bar_labels: bool = False,
    figsize: Optional[tuple] = None,
    label_fontsize: Optional[int] = None,
    advanced_time_dict: Optional[dict] = None,
    exclude_below_time_fraction: float = 1e-2,
    sort: bool = False,
):
    """
    Plots a detailed breakdown of execution times for a specific function.

    Parameters
    ----------
    function_name : str
        The name of the function to analyze.
    max_levels : int, optional
        The maximum levels of nested function calls to display, by default None.
    use_long_bar_labels : bool, optional
        Whether to use long labels for the bars, by default False.
    figsize : tuple, optional
        The figure size for the plot, by default None.
    label_fontsize: int, optional
        The font size for bar labels.
    advanced_time_dict : dict, optional
        By default, this will be the global `ADVANCED_TIME_DICT`. Override this
        to plot some other dict.
    exclude_below_time_fraction : float, optional
        Excludes functions contributing less than this fraction of total time, by default 1e-2.

    Returns
    -------
    dict
        A dictionary of execution times for each function in the call stack.
    """

    if advanced_time_dict is None:
        global ADVANCED_TIME_DICT
        advanced_time_dict = ADVANCED_TIME_DICT

    def find_key_in_nested_dict(nested_dict: dict, target_key):
        for key, value in nested_dict.items():
            if key == target_key:
                return value
            if isinstance(value, dict):
                result = find_key_in_nested_dict(value, target_key)
                if result is not None:
                    return result

    # Find the dictionary containing the function_name key
    result_dict = find_key_in_nested_dict(advanced_time_dict, function_name)
    if result_dict is None:
        raise ValueError(f"Key '{function_name}' not found in the nested dictionary.")

    result_dict = return_dict_subset_copy(result_dict, include, exclude)

    # Prepare data for plotting
    short_labels = []
    full_call_stack_labels = []
    times = []
    colors = []

    total_execution_time = result_dict["time"].sum()

    def collect_data(d, prefix="", function_name="", level=0):
        if max_levels is not None and level > max_levels:
            return
        symbol = " \u2192 "
        for sub_key, sub_value in d.items():
            if isinstance(sub_value, dict):
                collect_data(sub_value, prefix + sub_key + symbol, sub_key, level + 1)
            elif sub_key == "time":
                if sub_value.sum() / total_execution_time < exclude_below_time_fraction:
                    return
                if prefix == "":
                    full_call_stack_labels.append("Total")
                    short_labels.append("Total")
                    colors.append("dodgerblue")
                else:
                    full_call_stack_labels.append(prefix.rstrip("/")[: -len(symbol)])
                    short_labels.append(function_name)
                    colors.append("skyblue")
                times.append(sub_value.sum())

    # Get total execution times for each of the called functions
    collect_data({k: v for k, v in result_dict.items() if k != "time"})

    # Add the total execution time of the function at the front of the list
    times.insert(0, total_execution_time)
    full_call_stack_labels.insert(0, "Total")
    short_labels.insert(0, "Total")
    colors.insert(0, "dodgerblue")

    if use_long_bar_labels:
        bar_labels = full_call_stack_labels
    else:
        bar_labels = short_labels

    if top_n is not None:
        top_n = top_n + 1

    generate_time_barplot(
        times=times,
        labels=bar_labels,
        colors=colors,
        title=f"Execution time breakdown for\n{function_name}",
        figsize=figsize,
        label_fontsize=label_fontsize,
        sort=sort,
        top_n=top_n,
    )

    print(text_bf(f"Execution summary of {function_name}\n"))
    print(text_bf(f"Total execution time: {total_execution_time:.3g} s\n"))
    print(text_bf("Execution times:"))
    if sort:
        idx = np.argsort(times[1:]) + 1
    else:
        idx = range(1, len(short_labels))
    for i in idx:
        print(f"{i}. {short_labels[i]}: {times[i]:.2g} s")
    print()
    print(text_bf("Function call stack info:"))
    for i in range(1, len(full_call_stack_labels)):
        print(f"{i}. {full_call_stack_labels[i]}")

    selected_elapsed_times = dict(zip(short_labels, times))
    return selected_elapsed_times


def generate_time_barplot(
    times: dict,
    labels: list,
    colors: Optional[Union[list, str]],
    title: Optional[str] = None,
    figsize: Optional[tuple] = None,
    label_fontsize: Optional[int] = None,
    sort: bool = False,
    top_n: Optional[int] = None,
):
    """
    Generates a horizontal bar plot for elapsed times.

    Parameters
    ----------
    times : dict
        Elapsed times for the functions.
    labels : list
        Labels for the bars.
    colors : list or str, optional
        Colors for the bars, by default None.
    title : str, optional
        Title for the plot, by default None.
    figsize : tuple, optional
        Figure size for the plot, by default None.
    label_fontsize : int, optional
        Font size for the bar labels.
    """
    if label_fontsize is None:
        label_fontsize = default_label_font_size
    if sort:
        sort_idx = np.argsort(times)[::-1]
        times = [times[i] for i in sort_idx]
        labels = [labels[i] for i in sort_idx]
        if isinstance(colors, list):
            colors = [colors[i] for i in sort_idx]
    if top_n is not None:
        sort_idx = np.argsort(times)[::-1][:top_n]
        keep_idx = np.sort(sort_idx)
        times = [times[i] for i in keep_idx]
        labels = [labels[i] for i in keep_idx]
        if isinstance(colors, list):
            colors = [colors[i] for i in keep_idx]

    plt.figure(figsize=figsize)
    bars = plt.barh(range(len(times)), times, color=colors)
    plt.gca().invert_yaxis()  # Invert the y-axis to have the largest bar on top
    plt.tick_params(left=False, labelleft=False)

    # Add labels at the start of each bar
    for i, (bar, label) in enumerate(zip(bars, labels)):
        plt.text(
            0,  # Start at the left edge
            bar.get_y() + bar.get_height() / 2,  # Vertically centered
            " " + label,
            ha="left",
            va="center",
            fontsize=label_fontsize,
            fontweight="bold",
        )

    # Set x-axis and title
    plt.grid(linestyle=":")
    plt.gca().set_axisbelow(True)
    plt.xlabel("Elapsed Time (s)")
    plt.ylabel("Function")
    plt.title(title)
    # plt.tight_layout()
    plt.show()


def plot_elapsed_time_vs_call_number(
    elapsed_time_dict: Dict[str, List[float]],
    include: Optional[List[str]] = None,
    exclude: Optional[List[str]] = None,
    linestyle: str = "-",
    top_n: Optional[int] = None,  # plot top N largest sums
    exclude_below_time_fraction: float = 1e-2,
):
    elapsed_time_dict = return_dict_subset_copy(elapsed_time_dict, include, exclude)
    if top_n is not None:
        elapsed_time_dict = return_top_n_entries(elapsed_time_dict, top_n)

    total_execution_time = np.max([v.sum() for v in elapsed_time_dict.values()])

    for k, v in elapsed_time_dict.items():
        if hasattr(v, "__len__") and len(v) > 2:  # temp fix
            if v.sum() / total_execution_time > exclude_below_time_fraction:
                plt.plot(v, linestyle, label=k)

    # plt.legend(loc='upper left', bbox_to_anchor=(1.05, 1))
    plt.grid(linestyle=":")
    plt.gca().set_axisbelow(True)
    plt.ylabel("Elapsed Time (s)")
    plt.xlabel("Call number")
    plt.title("Elapsed time vs call number")
    plt.tight_layout()
    handles, labels = plt.gca().get_legend_handles_labels()
    plt.show()

    # Separate figure for the legend
    fig_legend = plt.figure()#figsize=(4, 2))  # Adjust size as needed
    fig_legend.legend(handles, labels, loc='center', frameon=False)
    plt.axis('off')  # Turn off the axes
    plt.show()


def return_dict_subset_copy(
    elapsed_time_dict: dict,
    include: Optional[List[str]] = None,
    exclude: Optional[List[str]] = None,
) -> dict:
    elapsed_time_dict = copy.deepcopy(elapsed_time_dict)
    if exclude is not None:
        [elapsed_time_dict.pop(k) for k in exclude]
    if include is not None:
        elapsed_time_dict = {
            k: elapsed_time_dict[k] for k in elapsed_time_dict.keys() if k in include
        }
    return elapsed_time_dict


def return_top_n_entries(elapsed_time_dict: dict, top_n: int) -> dict:
    elapsed_time_dict = copy.deepcopy(elapsed_time_dict)
    # Sort dictionary items based on the sum of their values and get top N
    sorted_items = sorted(
        elapsed_time_dict.items(),
        # key=lambda x: sum(x[1]) if hasattr(x[1], "__iter__") else float("-inf"),
        key=lambda x: sum(x[1]) if hasattr(x[1], "__iter__") else x[1],
        reverse=True,
    )[:top_n]
    keys = list(elapsed_time_dict.keys())
    keep_keys = list(dict(sorted_items).keys())
    for k in keys:
        if k not in keep_keys:
            elapsed_time_dict.pop(k)

    return elapsed_time_dict


def select_leaf_functions(advanced_time_dict: dict) -> dict[str, np.ndarray]:
    time_key = "time"

    def recursive_leaf_finder(
        nested_dict: dict, last_key: str = "top level", leaf_functions: dict = {}
    ) -> dict[str, np.ndarray]:
        # Check if function is a leaf by seeing if any of the
        # values in the dict are dicts
        is_a_leaf = np.any([isinstance(v, dict) for v in nested_dict.values()])
        if is_a_leaf:
            for k, v in nested_dict.items():
                if k != time_key:
                    recursive_leaf_finder(v, k, leaf_functions)
        else:
            if last_key in leaf_functions.keys():
                # Note: if a function is called in more than one
                # function then the array are no longer timer
                # ordered.
                leaf_functions[last_key] = np.append(
                    leaf_functions[last_key],
                    nested_dict[time_key],
                )
            else:
                leaf_functions[last_key] = nested_dict[time_key]
        return leaf_functions

    return recursive_leaf_finder(advanced_time_dict)


def text_bf(string: str) -> str:
    return f"\033[1m{string}\033[0m"



# Copyright © 2025 UChicago Argonne, LLC All right reserved
# Full license accessible at https://github.com//AdvancedPhotonSource/pty-chi/blob/main/LICENSE

from typing import TYPE_CHECKING

import torch
import numpy as np

import ptychi.position_correction as position_correction
import ptychi.data_structures.base as dsbase
import ptychi.image_proc as ip
import ptychi.maths as pmath
import ptychi.api.enums as enums
from ptychi.utils import to_numpy

if TYPE_CHECKING:
    import ptychi.api as api
    from ptychi.data_structures.probe import Probe
    from ptychi.data_structures.object import PlanarObject


class ProbePositions(dsbase.ReconstructParameter):
    options: "api.options.base.ProbePositionOptions"

    def __init__(
        self,
        *args,
        name: str = "probe_positions",
        options: "api.options.base.ProbePositionOptions" = None,
        **kwargs,
    ):
        """
        Probe positions.

        Parameters
        ----------
        data: a tensor of shape (N, 2) giving the probe positions in pixels.
            Input positions should be in row-major order, i.e., y-positions come first.
        """
        super().__init__(*args, name=name, options=options, is_complex=False, **kwargs)
        self.position_correction = position_correction.PositionCorrection(
            options=options.correction_options
        )
    
        self.register_buffer("initial_positions", self.data.detach())
        self.register_buffer(
            "affine_transform_matrix", 
            self.get_identity_affine_transform_matrix()
        )
        self.register_buffer("position_weights", torch.ones_like(self.data))
        self.affine_transform_components = {
            "scale": 1.0,
            "asymmetry": 0.0,
            "rotation": 0.0,
            "shear": 0.0,
        }

    @property
    def n_scan_points(self):
        return len(self.data)
    
    def get_identity_affine_transform_matrix(self, translation: bool = True):
        m = torch.eye(2, device=self.data.device)
        if translation:
            m = torch.cat([m, torch.zeros(2, 1, device=self.data.device)], dim=1)
        return m
    
    def get_slice_for_correction(self, n_slices: int = None):
        i_slice = self.options.correction_options.slice_for_correction
        if i_slice is None:
            if n_slices is None:
                raise ValueError(
                    "When `slice_for_correction` is not set, `n_slices` must "
                    "be provided to determine the middle slice."
                )
            i_slice = n_slices // 2
        return i_slice

    def get_positions_in_pixel(self):
        return self.data
    
    def position_mean_constraint_enabled(self, current_epoch: int):
        return self.options.constrain_position_mean and self.optimization_enabled(current_epoch)
    
    def constrain_position_mean(self):
        data = self.data
        data = data - data.mean(0)
        self.set_data(data)

    def update_position_weights(
        self, 
        probe: "Probe", 
        object_: "PlanarObject", 
        batch_size: int = 100,
    ):
        """Update the position weights based on patch-wise total variation.
        
        Parameters
        ----------
        probe: Probe
            The probe to use for the update.
        object_: PlanarObject
            The object to use for the update.
        batch_size: int
            The batch size for calculating patch-wise total variation. The computation
            is done in batches to avoid memory issues. This can, but does not have to
            be the same as the batch size used for reconstruction.
        """
        half_probe_shape = [x // 2 for x in probe.shape[-2:]]
        probe_intensity = ip.central_crop_or_pad(
            probe.data[0, 0].abs() ** 2, half_probe_shape
        )
        total_variations = torch.zeros_like(self.data)
        
        obj_slice = object_.get_slice(self.get_slice_for_correction(object_.n_slices))
        roi_bbox = object_.roi_bbox.get_bbox_with_top_left_origin()
        obj_slice = obj_slice / torch.max(obj_slice[roi_bbox.get_slicer()].abs())
        
        for ind_st in range(0, self.n_scan_points, batch_size):
            ind_end = min(ind_st + batch_size, self.n_scan_points)
            obj_patches = object_.extract_patches(
                self.data[ind_st:ind_end],
                patch_shape=half_probe_shape,
                integer_mode=True
            )
            obj_patches = obj_patches[:, self.get_slice_for_correction(object_.n_slices)]
            obj_patches = ip.vignette(
                obj_patches, margin=int(np.mean(half_probe_shape) * 0.15), sigma=np.mean(half_probe_shape) * 0.05, method="gaussian"
            )
            grads_y, grads_x = ip.fourier_gradient(obj_patches)
            
            illumination_map = object_.preconditioner
            illumination_map = ip.gaussian_filter(illumination_map, sigma=np.mean(probe.shape[-2:]) * 0.1)
            illum_patches = ip.extract_patches_integer(
                object_.preconditioner, 
                self.data[ind_st:ind_end] + object_.pos_origin_coords,
                half_probe_shape
            )
            
            grads_y = grads_y.abs() * illum_patches * probe_intensity
            grads_x = grads_x.abs() * illum_patches * probe_intensity
            total_variations[ind_st:ind_end] = torch.sqrt(torch.stack([
                    torch.mean(grads_y, dim=(-2, -1)),
                    torch.mean(grads_x, dim=(-2, -1))
                ], dim=-1)
            )
        
        self.position_weights = total_variations ** 4
        self.position_weights = self.position_weights / self.position_weights.mean()
        
    def update_affine_transform_matrix(self):
        """Fit the affine transformation matrix that relates the initial positions to 
        the current positions.
        """
        dofs = self.options.affine_transform_constraint.degrees_of_freedom
        x = self.position_weights * self.data
        x0 = self.position_weights * self.initial_positions
        
        if enums.AffineDegreesOfFreedom.TRANSLATION in dofs:
            x = torch.cat([x, torch.ones_like(x[..., 0:1])], dim=-1)
            x0 = torch.cat([x0, torch.ones_like(x0[..., 0:1])], dim=-1)
        
        # `fit_linear_transform_matrix(x0, x)` finds A in x0 A = x. When we apply affine transform,
        # we do x = (A x0^T)^T = x0 A^T. Therefore, we need to take the transpose of the result.
        a_mat = pmath.fit_linear_transform_matrix(x0, x).T
        
        if enums.AffineDegreesOfFreedom.TRANSLATION in dofs:
            a_mat = a_mat[:, :-1]
        
        scale, asymmetry, rotation, shear = pmath.decompose_2x2_affine_transform_matrix(a_mat)
        if enums.AffineDegreesOfFreedom.SCALE not in dofs:
            scale = 1.0
        if enums.AffineDegreesOfFreedom.ASYMMETRY not in dofs:
            asymmetry = 0.0
        if enums.AffineDegreesOfFreedom.ROTATION not in dofs:
            rotation = 0.0
        if enums.AffineDegreesOfFreedom.SHEAR not in dofs:
            shear = 0.0
        
        a_mat = pmath.compose_2x2_affine_transform_matrix(
            scale, asymmetry, rotation, shear
        )
        self.affine_transform_matrix = torch.cat([a_mat, torch.zeros(2, 1, device=a_mat.device)], dim=1)
        
        # Update saved components.
        self.affine_transform_components["scale"] = to_numpy(scale)
        self.affine_transform_components["asymmetry"] = to_numpy(asymmetry)
        self.affine_transform_components["rotation"] = to_numpy(rotation)
        self.affine_transform_components["shear"] = to_numpy(shear)
        
    def apply_affine_transform_constraint(self):
        estimated_positions = self.affine_transform_matrix @ \
            torch.cat([self.initial_positions, torch.ones_like(self.initial_positions[..., 0:1])], dim=-1).T
        estimated_positions = estimated_positions.T
        residuals = self.data - estimated_positions
        
        residuals = residuals - residuals.mean(0)
        errors = residuals.abs()
        max_error = self.options.affine_transform_constraint.max_expected_error
        
        relax = 0.1
        flexibility = relax / (1 + self.position_weights)
        flexibility = torch.clip(flexibility + (errors - max_error).clip(min=0) ** 2 / max_error ** 2, max=10 * relax)
        
        pos_new = self.data * (1 - flexibility) + flexibility * estimated_positions
        self.set_data(pos_new)

    def step_optimizer(self, *args, **kwargs):
        """Step the optimizer with gradient filled in. This function
        can optionally impose a limit on the magnitude of the update.
        """
        limit_user = self.options.correction_options.update_magnitude_limit
        if limit_user is not None and limit_user <= 0:
            raise ValueError("`update_magnitude_limit` should either be None or a positive number.")
        if limit_user == torch.inf:
            limit_user = None
        data0 = self.data
            
        self.optimizer.step()
        
        if not self.options.correction_options.clip_update_magnitude_by_mad and limit_user is None:
            return
        
        # Get update.
        data = self.data
        dx = data - data0
        update_mag = dx.abs()
        update_signs = dx.sign()
        
        # Truncate update by the limit and reapply it.
        if self.options.correction_options.clip_update_magnitude_by_mad:
            limit_mad = pmath.mad(dx, dim=0) * 10
        else:
            limit_mad = torch.inf
        if limit_user is not None:
            limit = torch.clip(limit_mad, max=limit_user)
        else:
            limit = limit_mad
        dx = update_mag.clip(max=limit) * update_signs
        
        with torch.no_grad():
            self.set_data(data0 + dx)



# Copyright © 2025 UChicago Argonne, LLC All right reserved
# Full license accessible at https://github.com//AdvancedPhotonSource/pty-chi/blob/main/LICENSE

from __future__ import annotations
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import TypeAlias
import cmath
import math

import torch
from torch.fft import fftfreq

from ptychi.timing.timer_utils import timer
import ptychi.maths as pmath
import ptychi.utils as utils


BooleanTensor: TypeAlias = torch.Tensor
ComplexTensor: TypeAlias = torch.Tensor
RealTensor: TypeAlias = torch.Tensor


@dataclass(frozen=True)
class WavefieldPropagatorParameters:
    width_px: int
    """number of pixels in the x-direction"""
    height_px: int
    """number of pixels in the y-direction"""
    wavelength_m: float
    """illumination wavelength in meters"""
    pixel_width_wlu: float
    """pixel width in wavelengths"""
    pixel_aspect_ratio: float
    """pixel aspect ratio (width / height)"""
    propagation_distance_wlu: float
    """propagation distance in wavelengths"""

    @classmethod
    def create_simple(
        cls,
        wavelength_m: float,
        width_px: int,
        height_px: int,
        pixel_width_m: float,
        pixel_height_m: float,
        propagation_distance_m: float,
    ) -> WavefieldPropagatorParameters:
        """
        Creates propagator paramaters dataclass from quantities with physical length units.

        Parameters
        ----------
        wavelength_m : float
            Illumination wavelength in meters
        width_px : int
            Number of pixels in the x-direction
        height_px : int
            Number of pixels in the y-direction
        pixel_width_m : float
            Source plane pixel width in meters
        pixel_height_m : float
            Source plane pixel height in meters
        propagation_distance_m : float
            Propagation distance in meters

        Returns
        -------
        WavefieldPropagatorParameters
            a dataclass that contains the nondimensionalized propagator parameters
        """
        return cls(
            width_px=width_px,
            height_px=height_px,
            wavelength_m=wavelength_m,
            pixel_width_wlu=pixel_width_m / wavelength_m,
            pixel_aspect_ratio=pixel_width_m / pixel_height_m,
            propagation_distance_wlu=propagation_distance_m / wavelength_m,
        )

    @property
    def fresnel_number(self) -> float:
        pixel_width_wlu_sq = self.pixel_width_wlu * self.pixel_width_wlu
        return pixel_width_wlu_sq / self.propagation_distance_wlu
    
    @property
    def pixel_width_m(self) -> float:
        return self.pixel_width_wlu * self.wavelength_m
    
    @property
    def pixel_height_m(self) -> float:
        return self.pixel_width_m / self.pixel_aspect_ratio
    
    @property
    def pixel_height_wlu(self) -> float:
        return self.pixel_width_wlu / self.pixel_aspect_ratio
    
    @property
    def propagation_distance_m(self) -> float:
        return self.propagation_distance_wlu * self.wavelength_m

    def get_spatial_coordinates(self) -> tuple[RealTensor, RealTensor]:
        ii = torch.arange(self.width_px)
        jj = torch.arange(self.height_px)
        JJ, II = torch.meshgrid(jj, ii, indexing="ij")
        XX = II - self.width_px // 2
        YY = JJ - self.height_px // 2
        XX = utils.auto_transfer_to_device(XX)
        YY = utils.auto_transfer_to_device(YY)
        return YY, XX

    def get_frequency_coordinates(self) -> tuple[RealTensor, RealTensor]:
        fx = fftfreq(self.width_px)
        fy = fftfreq(self.height_px)
        FY, FX = torch.meshgrid(fy, fx, indexing="ij")
        FY = utils.auto_transfer_to_device(FY)
        FX = utils.auto_transfer_to_device(FX)
        return FY.detach(), FX.detach()
    
    def is_fresnel_transform_preferrable(self) -> bool:
        n_eff = math.sqrt(float(self.height_px) * float(self.width_px))
        return self.fresnel_number < 1 / (2 * n_eff)


class WavefieldPropagator(ABC, torch.nn.Module):
    @abstractmethod
    def propagate_forward(self, wavefield: ComplexTensor) -> ComplexTensor:
        pass

    @abstractmethod
    def propagate_backward(self, wavefield: ComplexTensor) -> ComplexTensor:
        pass

    def forward(self, wavefield: ComplexTensor) -> ComplexTensor:
        return self.propagate_forward(wavefield)


class FourierPropagator(WavefieldPropagator):
    def __init__(self, norm=None) -> None:
        super().__init__()
        self.norm = norm

    @timer()
    def propagate_forward(self, wavefield: ComplexTensor) -> ComplexTensor:
        return pmath.fft2_precise(wavefield, norm=self.norm)

    @timer()
    def propagate_backward(self, wavefield: ComplexTensor) -> ComplexTensor:
        return pmath.ifft2_precise(wavefield, norm=self.norm)

class AngularSpectrumPropagator(WavefieldPropagator):
    def __init__(self, parameters: WavefieldPropagatorParameters) -> None:
        super().__init__()

        _transfer_function = self.get_transfer_function(parameters)
        
        # Separate registered buffer into real and imaginary parts to prevent it
        # from breaking in DataParallel.
        self._transfer_function_real = _transfer_function.real
        self._transfer_function_imag = _transfer_function.imag

    def update(self, parameters: WavefieldPropagatorParameters) -> None:
        _transfer_function = self.get_transfer_function(parameters)
        self._transfer_function_real = _transfer_function.real
        self._transfer_function_imag = _transfer_function.imag

    def get_transfer_function(self, parameters: WavefieldPropagatorParameters) -> ComplexTensor:
        ar = parameters.pixel_aspect_ratio

        i2piz = 2j * torch.pi * parameters.propagation_distance_wlu

        FY, FX = parameters.get_frequency_coordinates()
        FY, FX = FY.double(), FX.double()
        F2 = torch.square(FX) + torch.square(ar * FY)
        self.register_buffer('F2', F2)

        ratio = self.F2 / (parameters.pixel_width_wlu**2)
        tf = torch.exp(i2piz * torch.sqrt(1 - ratio))
        tf = torch.where(ratio < 1, tf, 1)
        tf = tf.to(utils.get_default_complex_dtype())
        return tf

    def propagate_forward(self, wavefield: ComplexTensor) -> ComplexTensor:
        tf = self._transfer_function_real + 1j * self._transfer_function_imag
        return pmath.ifft2_precise(tf * pmath.fft2_precise(wavefield))

    def propagate_backward(self, wavefield: ComplexTensor) -> ComplexTensor:
        tf = self._transfer_function_real + 1j * self._transfer_function_imag
        return pmath.ifft2_precise(pmath.fft2_precise(wavefield) / tf)


class FresnelTransformPropagator(WavefieldPropagator):
    def __init__(self, parameters: WavefieldPropagatorParameters) -> None:
        super().__init__()

        _C0, _C1C2, _B = self.get_kernels(parameters)
        self._C0 = _C0
        self.register_buffer('_C1C2', _C1C2)
        self.register_buffer('_B', _B)
        
    def get_kernels(
        self, 
        parameters: WavefieldPropagatorParameters
    ) -> tuple[ComplexTensor, ComplexTensor, ComplexTensor]:
        ipi = 1j * torch.pi
        Fr = float(parameters.fresnel_number)
        ar = float(parameters.pixel_aspect_ratio)
        N = float(parameters.width_px)
        M = float(parameters.height_px)
        YY, XX = parameters.get_spatial_coordinates()
        YY, XX = YY.double(), XX.double()
        
        C0 = Fr / (1j * ar)
        C1 = cmath.exp(2j * cmath.pi * parameters.propagation_distance_wlu)
        C2 = torch.exp((torch.square(XX / N) + torch.square(ar * YY / M)) * ipi / Fr)
        C1C2 = C1 * C2
        B = torch.exp(ipi * Fr * (torch.square(XX) + torch.square(YY / ar)))
        
        C1C2 = C1C2.to(utils.get_default_complex_dtype())
        B = B.to(utils.get_default_complex_dtype())
        return C0, C1C2, B

    def propagate_forward(self, wavefield: ComplexTensor) -> ComplexTensor:
        A = self._C1C2 * self._C0
        return (A * pmath.fft2_precise(wavefield * self._B)).to(utils.get_default_complex_dtype())

    def propagate_backward(self, wavefield: ComplexTensor) -> ComplexTensor:
        A = self._C1C2 / self._C0
        return (self._B * pmath.ifft2_precise(wavefield * A)).to(utils.get_default_complex_dtype())


class FraunhoferPropagator(WavefieldPropagator):
    def __init__(self, parameters: WavefieldPropagatorParameters) -> None:
        super().__init__()
        ipi = 1j * torch.pi

        Fr = parameters.fresnel_number
        ar = parameters.pixel_aspect_ratio
        N = parameters.width_px
        M = parameters.height_px
        YY, XX = parameters.get_spatial_coordinates()

        self._C0 = Fr / (1j * ar)
        self._C1 = cmath.exp(2j * cmath.pi * parameters.propagation_distance_wlu)
        self._C2 = torch.exp((torch.square(XX / N) + torch.square(ar * YY / M)) * ipi / Fr)

    def propagate_forward(self, wavefield: ComplexTensor) -> ComplexTensor:
        A = self._C2 * self._C1 * self._C0
        return A * pmath.fft2_precise(wavefield)

    def propagate_backward(self, wavefield: ComplexTensor) -> ComplexTensor:
        A = self._C2 * self._C1 / self._C0
        return pmath.ifft2_precise(wavefield * A)



# Copyright © 2025 UChicago Argonne, LLC All right reserved
# Full license accessible at https://github.com//AdvancedPhotonSource/pty-chi/blob/main/LICENSE

from typing import Optional, Literal

import torch
from torch import Tensor


class MSELossOfSqrt(torch.nn.MSELoss):
    def __init__(self, eps=1e-7, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.eps = eps

    def forward(
        self, y_pred: Tensor, y_true: Tensor, weight_map: Optional[Tensor] = None
    ) -> Tensor:
        if weight_map is None:
            loss = super().forward(torch.sqrt(y_pred + self.eps), torch.sqrt(y_true + self.eps))
        else:
            d = ((torch.sqrt(y_pred + self.eps) - torch.sqrt(y_true + self.eps)) ** 2) * weight_map

            if self.reduction == "mean":
                loss = torch.mean(d)
            elif self.reduction == "sum":
                loss = torch.sum(d)

        return loss


class TotalVariationLoss(torch.nn.Module):
    def __init__(self, reduction: Literal["mean", "sum"] = "mean", *args, **kwargs) -> None:
        """Calculates the total variation of one or more 2D
        images along the last two dimensions.
        """
        super().__init__(*args, **kwargs)
        self.reduction = reduction

    def forward(self, img: Tensor) -> Tensor:
        """The forward pass.

        Parameters
        ----------
        img : Tensor
            A (..., H, W) tensor.

        Returns
        -------
        Tensor
            The total variation of the input image(s).
        """
        tv = torch.norm(img[..., :, 1:] - img[..., :, :-1], p=2, dim=(-2, -1)) + torch.norm(
            img[..., 1:, :] - img[..., :-1, :], p=2, dim=(-2, -1)
        )
        if self.reduction == "mean":
            return torch.mean(tv)
        elif self.reduction == "sum":
            return torch.sum(tv)



# Copyright © 2025 UChicago Argonne, LLC All right reserved
# Full license accessible at https://github.com//AdvancedPhotonSource/pty-chi/blob/main/LICENSE

from typing import Optional, TYPE_CHECKING

import torch
from torch.utils.data import Dataset


import ptychi.forward_models as fm
from ptychi.reconstructors.base import IterativeReconstructor, LossTracker
import ptychi.maps as maps
if TYPE_CHECKING:
    import ptychi.data_structures.parameter_group as pg
    import ptychi.api as api


class AutodiffReconstructor(IterativeReconstructor):
    def __init__(
        self,
        parameter_group: "pg.ParameterGroup",
        dataset: Dataset,
        options: Optional["api.options.ad_general.AutodiffReconstructorOptions"] = None,
        *args,
        **kwargs,
    ) -> None:
        super().__init__(
            parameter_group=parameter_group,
            dataset=dataset,
            options=options,
            *args,
            **kwargs,
        )
        self.forward_model_class = options.forward_model_class
        if not isinstance(self.forward_model_class, type):
            self.forward_model_class = maps.get_forward_model_by_enum(self.forward_model_class)
        self.forward_model_params = options.forward_model_params if options.forward_model_params is not None else {}
        self.forward_model = None
        self.loss_function = maps.get_loss_function_by_enum(options.loss_function)()

    def build(self) -> None:
        super().build()
        self.build_forward_model()

    def build_loss_tracker(self):
        f = self.loss_function if self.displayed_loss_function is None else self.displayed_loss_function
        # LossTracker should always compute the loss using data if metric function and loss function
        # are not the same type.
        always_compute_loss = (self.displayed_loss_function is not None) and (
            type(self.displayed_loss_function) is not type(self.loss_function)
        )
        self.loss_tracker = LossTracker(metric_function=f, always_compute_loss=always_compute_loss)

    def build_forward_model(self):
        self.forward_model = self.forward_model_class(
            self.parameter_group, **self.forward_model_params
        )
        if not torch.get_default_device().type == "cpu":
            self.forward_model = torch.nn.DataParallel(self.forward_model)
            self.forward_model.to(torch.get_default_device())

    def run_post_differentiation_hooks(self, input_data, y_true):
        self.get_forward_model().post_differentiation_hook(*input_data, y_true)

    def apply_regularizers(self) -> None:
        """
        Apply Tikonov regularizers.
        """
        reg = torch.tensor(0.0)
        return reg

    def run_minibatch(self, input_data, y_true, *args, **kwargs):
        y_pred = self.forward_model(*input_data)
        batch_loss = self.loss_function(y_pred, y_true)

        batch_loss.backward()
        self.run_post_differentiation_hooks(input_data, y_true)
        reg_loss = self.apply_regularizers()
        self.step_all_optimizers()
        self.forward_model.zero_grad()
        self.run_post_update_hooks()

        self.loss_tracker.update_batch_loss(y_pred=y_pred, y_true=y_true, loss=batch_loss.item())
        self.loss_tracker.update_batch_regularization_loss(reg_loss.item())

    def step_all_optimizers(self):
        for var in self.parameter_group.get_optimizable_parameters():
            if var.optimization_enabled(self.current_epoch):
                var.step_optimizer()
                for sub_module in var.optimizable_sub_modules:
                    if sub_module.optimization_enabled(self.current_epoch):
                        sub_module.step_optimizer()

    def get_forward_model(self) -> "fm.ForwardModel":
        if isinstance(self.forward_model, torch.nn.DataParallel):
            return self.forward_model.module
        else:
            return self.forward_model

    def get_config_dict(self) -> dict:
        d = super().get_config_dict()
        d.update(
            {
                "forward_model_class": str(self.forward_model_class),
                "loss_function": str(self.loss_function),
            }
        )
        return d



# Copyright © 2025 UChicago Argonne, LLC All right reserved
# Full license accessible at https://github.com//AdvancedPhotonSource/pty-chi/blob/main/LICENSE

from typing import Optional, Union, TYPE_CHECKING
import logging
import enum
import os

import numpy as np
import tifffile
import torch
from torch import Tensor

import ptychi.image_proc as ip
import ptychi.maths as pmath
import ptychi.utils as utils
import ptychi.data_structures.base as dsbase
import ptychi.data_structures.object as object
import ptychi.data_structures.opr_mode_weights as oprweights
from ptychi.propagate import FourierPropagator, WavefieldPropagator
import ptychi.maps as maps

if TYPE_CHECKING:
    import ptychi.api as api

logger = logging.getLogger(__name__)


class ProbeRepresentation(enum.StrEnum):
    NORMAL = enum.auto()
    SPARSE_CODE = enum.auto()
    DIP = enum.auto()


class Probe(dsbase.ReconstructParameter):
    # TODO: eigenmode_update_relaxation is only used for LSQML. We should create dataclasses
    # to contain additional options for ReconstructParameter classes, and subclass them for specific
    # reconstruction algorithms - for example, ProbeOptions -> LSQMLProbeOptions.
    options: "api.options.base.ProbeOptions"
    
    representation: ProbeRepresentation = ProbeRepresentation.NORMAL

    def __init__(
        self,
        name: str = "probe",
        options: "api.options.base.ProbeOptions" = None,
        *args,
        **kwargs,
    ):
        """
        Represents the probe function in a tensor of shape
            `(n_opr_modes, n_modes, h, w)`
        where:
        - n_opr_modes is the number of mutually coherent probe modes used in orthogonal
          probe relaxation (OPR).
        - n_modes is the number of mutually incoherent probe modes.
        """
        super().__init__(*args, name=name, options=options, is_complex=True, **kwargs)
        if len(self.shape) != 4:
            raise ValueError("Probe tensor must be of shape (n_opr_modes, n_modes, h, w).")

        self.probe_power = options.power_constraint.probe_power
        self.orthogonalize_incoherent_modes_method = options.orthogonalize_incoherent_modes.method

    def shift(self, shifts: Tensor):
        """
        Generate shifted probe.

        Parameters
        ----------
        shifts : Tensor
            A tensor of shape (2,) or (N, 2) giving the shifts in pixels.
            If a (N, 2)-shaped tensor is given, a batch of shifted probes are generated.

        Returns
        -------
        shifted_probe : Tensor
            Shifted probe.
        """
        if shifts.ndim == 1:
            probe_straightened = self.tensor.complex().view(-1, *self.shape[-2:])
            shifted_probe = ip.fourier_shift(
                probe_straightened, shifts[None, :].repeat([probe_straightened.shape[0], 1])
            )
            shifted_probe = shifted_probe.view(*self.shape)
        else:
            n_shifts = shifts.shape[0]
            n_images_each_probe = self.shape[0] * self.shape[1]
            probe_straightened = self.tensor.complex().view(n_images_each_probe, *self.shape[-2:])
            probe_straightened = probe_straightened.repeat(n_shifts, 1, 1)
            shifts = shifts.repeat_interleave(n_images_each_probe, dim=0)
            shifted_probe = ip.fourier_shift(probe_straightened, shifts)
            shifted_probe = shifted_probe.reshape(n_shifts, *self.shape)
        return shifted_probe

    @property
    def n_modes(self):
        return self.tensor.shape[1]

    @property
    def n_opr_modes(self):
        return self.tensor.shape[0]

    @property
    def has_multiple_opr_modes(self):
        return self.n_opr_modes > 1
    
    @property
    def lateral_shape(self):
        return self.shape[-2:]

    @property
    def has_multiple_incoherent_modes(self):
        return self.n_modes > 1

    def get_mode(self, mode: int, keepdim: bool = False):
        mode = self.tensor.complex()[:, mode]
        if keepdim:
            mode = mode.unsqueeze(1)
        return mode

    def get_opr_mode(self, mode: int, keepdim: bool = False):
        mode = self.tensor.complex()[mode]
        if keepdim:
            mode = mode.unsqueeze(0)
        return mode

    def get_mode_and_opr_mode(self, mode: int, opr_mode: int):
        return self.tensor.complex()[opr_mode, mode]

    def get_spatial_shape(self):
        return self.shape[-2:]

    def get_all_mode_intensity(
        self,
        opr_mode: Optional[int] = 0,
        weights: Optional[Union[Tensor, "dsbase.ReconstructParameter"]] = None,
    ) -> Tensor:
        """
        Get the intensity of all probe modes.

        Parameters
        ----------
        opr_mode : Optional[int]
            the OPR mode. If this is not None, only the intensity of the chosen
            OPR mode is calculated. Otherwise, it calculates the intensity of the weighted sum
            of all OPR modes. In that case, `weights` must be given.
        weights : Optional[Union[Tensor, ReconstructParameter]]
            a (n_opr_modes,) tensor giving the weights of OPR modes.

        Returns
        -------
        intensity : Tensor
            The summed intensity of all probe modes.
        """
        if isinstance(weights, oprweights.OPRModeWeights):
            weights = weights.data
        if opr_mode is not None:
            p = self.data[opr_mode]
        else:
            p = (self.data * weights[None, :, :, :]).sum(0)
        return torch.sum((p.abs()) ** 2, dim=0)

    def get_unique_probes(
        self, weights: Union[Tensor, "dsbase.ReconstructParameter"], mode_to_apply: Optional[int] = None
    ) -> Tensor:
        """
        Parameters
        ----------
        weights : Tensor
            A tensor giving the weights of the eigenmodes.
        mode_to_apply : int, optional
            The incoherent mode for which OPR modes should be applied. The data
            for other modes will be set to the value of the first OPR mode. If None,
            OPR correction will be done to all incoherent modes.

        Returns
        -------
        unique_probe : Tensor
            A tensor of unique probes. If weights.ndim == 2, the shape is (n_points, n_modes, h, w).
            If weights.ndim == 1, the shape is (n_modes, h, w).
        """
        if isinstance(weights, oprweights.OPRModeWeights):
            weights = weights.data

        p_orig = None
        if mode_to_apply is not None:
            p_orig = self.data.clone()
            p = p_orig[:, [mode_to_apply], :, :]
        else:
            p = self.data.clone()
        if weights.ndim == 1:
            unique_probe = p * weights[:, None, None, None]
            unique_probe = unique_probe.sum(0)
        else:
            unique_probe = p[None, ...] * weights[:, :, None, None, None]
            unique_probe = unique_probe.sum(1)

        # If OPR is only applied on one incoherent mode, add in the rest of the modes.
        if mode_to_apply is not None:
            if weights.ndim == 1:
                # Shape of unique_probe:     (1, h, w)
                p_orig[0, [mode_to_apply], :, :] = unique_probe
                unique_probe = p_orig[0, ...]
            else:
                # Shape of unique_probe:     (n_points, 1, h, w)
                p_orig = p_orig[0:1, ...].repeat(weights.shape[0], 1, 1, 1)
                p_orig[:, [mode_to_apply], :, :] = unique_probe
                unique_probe = p_orig
        return unique_probe

    def _get_probe_mode_slicer(self, mode_index=None):
        if mode_index is None:
            return slice(None)
        else:
            return slice(mode_index, mode_index + 1)

    def constrain_incoherent_modes_orthogonality(self):
        """Orthogonalize the incoherent probe modes for the first OPR mode."""
        if not self.has_multiple_incoherent_modes:
            return

        probe = self.data

        norm_first_mode_orig = pmath.norm(probe[0, 0], dim=(-2, -1))

        if self.orthogonalize_incoherent_modes_method == "gs":
            func = pmath.orthogonalize_gs
        elif self.orthogonalize_incoherent_modes_method == "svd":
            func = pmath.orthogonalize_svd
        else:
            raise NotImplementedError(
                f"Orthogonalization method {self.orthogonalize_incoherent_modes_method} "
                "is not supported."
            )
        probe[0] = func(
            probe[0],
            dim=(-2, -1),
            group_dim=0,
        )

        # Restore norm.
        if self.orthogonalize_incoherent_modes_method != "svd":
            norm_first_mode_new = pmath.norm(probe[0, 0], dim=(-2, -1))
            probe = probe * norm_first_mode_orig / norm_first_mode_new

        self.set_data(probe)

    def constrain_opr_mode_orthogonality(
        self, weights: "oprweights.OPRModeWeights", eps=1e-5, *, update_weights_in_place: bool
    ):
        """
        Add the following constraints to variable probe weights

        1. Remove outliars from weights
        2. Enforce orthogonality once per epoch
        3. Sort the variable probes by their total energy
        4. Normalize the variable probes so the energy is contained in the weight

        Adapted from Tike (https://github.com/AdvancedPhotonSource/tike). The implementation
        in Tike assumes a separately stored variable probe eigenmodes; here we use the
        PtychoShelves convention and regard the second and following OPR modes as eigenmodes.

        Also, this function assumes that OPR correction is only applied to the first
        incoherent mode when mixed state probe is used, as this is what PtychoShelves does.
        OPR modes of other incoherent modes are ignored, for now.

        This function also updates the OPR mode weights when `update_weights_in_place` is True.

        Parameters
        ----------
        weights : OPRModeWeights
            The OPR mode weights object.
        update_weights_in_place : bool
            Whether to update the OPR mode weights data.

        Returns
        -------
        Tensor, optional
            Normalized and sorted OPR mode weights.
        """
        if not self.has_multiple_opr_modes:
            return

        weights_data = weights.data

        # The main mode of the probe is the first OPR mode, while the
        # variable part of the probe is the second and following OPR modes.
        # The main mode should not change during orthogonalization, but the
        # variable OPR modes should all be orthogonal to it.
        probe = self.data

        # TODO: remove outliars by polynomial fitting (remove_variable_probe_ambiguities.m)

        # Normalize eigenmodes and adjust weights.
        eigenmodes = probe[1:, ...]
        vnorm = pmath.mnorm(eigenmodes, dim=(-2, -1), keepdims=True)
        eigenmodes /= vnorm + eps
        # Shape of weights:      (n_points, n_opr_modes).
        # Currently, only the first incoherent mode has OPR modes, and the
        # stored weights are for that mode.
        weights_data[:, 1:] = weights_data[:, 1:] * vnorm[:, 0, 0, 0]

        # Orthogonalize variable probes. With Gram-Schmidt, the first
        # OPR mode (i.e., the main mode) should not change during orthogonalization.
        probe = pmath.orthogonalize_gs(
            probe,
            dim=(-2, -1),
            group_dim=0,
        )

        if False:
            # Compute the energies of variable OPR modes (i.e., the second and following)
            # in order to sort probes by energy.
            # Shape of power:         (n_opr_modes - 1,).
            power = pmath.norm(weights_data[..., 1:], dim=0) ** 2

            # Sort the probes by energy
            sorted = torch.argsort(-power)
            weights_data[:, 1:] = weights_data[:, sorted + 1]
            # Apply only to the first incoherent mode.
            probe[1:, 0, :, :] = probe[sorted + 1, 0, :, :]

        # Remove outliars from variable probe weights.
        aevol = torch.abs(weights_data)
        weights_data = torch.minimum(
            aevol,
            1.5
            * torch.quantile(
                aevol,
                0.95,
                dim=0,
                keepdims=True,
            ).type(weights_data.dtype),
        ) * torch.sign(weights_data)

        # Update stored data.
        self.set_data(probe)

        if update_weights_in_place:
            weights.set_data(weights_data)
        else:
            return weights_data

    def constrain_probe_power(
        self,
        object_: "object.Object",
        opr_mode_weights: Union[Tensor, "oprweights.OPRModeWeights"],
        propagator: Optional[WavefieldPropagator] = None,
    ) -> None:
        if self.probe_power <= 0.0:
            return

        if isinstance(opr_mode_weights, oprweights.OPRModeWeights):
            opr_mode_weights = opr_mode_weights.data

        if propagator is None:
            propagator = FourierPropagator()

        # Shape of probe_composed:        (n_modes, h, w)
        if self.has_multiple_opr_modes:
            avg_weights = opr_mode_weights.mean(dim=0)
            probe_composed = self.get_unique_probes(avg_weights, mode_to_apply=0)
        else:
            probe_composed = self.get_opr_mode(0)

        # TODO: use propagator for forward simulation
        propagated_probe = propagator.propagate_forward(probe_composed)
        propagated_probe_power = torch.sum(propagated_probe.abs() ** 2)
        power_correction = torch.sqrt(self.probe_power / propagated_probe_power)

        self.set_data(self.data * power_correction)
        object_.set_data(object_.data / power_correction)

        logger.info("Probe and object scaled by {}.".format(power_correction))

    def constrain_support(self):
        """
        Apply probe support constraint through a mask generated by thresholding the
        blurred probe magnitude.
        """
        data = self.data
        mask = ip.gaussian_filter(data, sigma=3, size=5).abs()
        thresh = (
            mask.max(-1, keepdim=True).values.max(-2, keepdim=True).values
            * self.options.support_constraint.threshold
        )
        mask = torch.where(mask > thresh, 1.0, 0.0)
        mask = ip.gaussian_filter(mask, sigma=2, size=3).abs()
        data = data * mask
        self.set_data(data)

    def center_probe(self):
        """
        Move the probe's center of mass to the center of the probe array.
        """
        com = ip.find_center_of_mass(self.get_mode_and_opr_mode(0, 0))
        shift = utils.to_tensor(self.shape[-2:]) // 2 - com
        shifted_probe = self.shift(shift)
        self.set_data(shifted_probe)

    def post_update_hook(self) -> None:
        super().post_update_hook()

    def normalize_eigenmodes(self):
        """
        Normalize all eigenmodes (the second and following OPR modes) such that each of them
        has a squared norm equal to the number of pixels in the probe.
        """
        if not self.has_multiple_opr_modes:
            return
        eigen_modes = self.data[1:, ...]
        for i_opr_mode in range(eigen_modes.shape[0]):
            for i_mode in range(eigen_modes.shape[1]):
                eigen_modes[i_opr_mode, i_mode, :, :] /= (
                    pmath.mnorm(eigen_modes[i_opr_mode, i_mode, :, :], dim=(-2, -1)) + 1e-8
                )

        new_data = self.data
        new_data[1:, ...] = eigen_modes
        self.set_data(new_data)

    def save_tiff(self, path: str):
        """
        Save the probe's magnitude and phase as 2 TIFF files. Each file contains
        an array of tiles, where the rows correspond to incoherent probe modes
        and columns correspond to OPR modes.

        Parameters
        ----------
        path : str
            Path to save. "_phase" and "_mag" will be appended to the filename.
        """
        fname = os.path.splitext(path)[0]
        mag_img = np.empty([self.shape[3] * self.shape[1], self.shape[2] * self.shape[0]])
        phase_img = np.empty([self.shape[3] * self.shape[1], self.shape[2] * self.shape[0]])
        data = self.data
        for i_mode in range(self.shape[1]):
            for i_opr_mode in range(self.shape[0]):
                mag_img[
                    i_mode * self.shape[3] : (i_mode + 1) * self.shape[3],
                    i_opr_mode * self.shape[2] : (i_opr_mode + 1) * self.shape[2],
                ] = data[i_opr_mode, i_mode, :, :].abs().detach().cpu().numpy()
                phase_img[
                    i_mode * self.shape[3] : (i_mode + 1) * self.shape[3],
                    i_opr_mode * self.shape[2] : (i_opr_mode + 1) * self.shape[2],
                ] = torch.angle(data[i_opr_mode, i_mode, :, :]).detach().cpu().numpy()
        tifffile.imwrite(fname + "_mag.tif", mag_img)
        tifffile.imwrite(fname + "_phase.tif", phase_img)

class SynthesisDictLearnProbe( Probe ):
    
    representation: ProbeRepresentation = ProbeRepresentation.SPARSE_CODE
    
    def __init__(self, name = "probe", options = None, *args, **kwargs):    
        
        super().__init__(name, options, build_optimizer=False, data_as_parameter=False, *args, **kwargs)

        dictionary_matrix, dictionary_matrix_pinv, dictionary_matrix_H = self.get_dictionary()
        self.register_buffer("dictionary_matrix", dictionary_matrix)
        self.register_buffer("dictionary_matrix_pinv", dictionary_matrix_pinv)
        self.register_buffer("dictionary_matrix_H", dictionary_matrix_H)
        
        probe_sparse_code_nnz = torch.tensor(  self.options.experimental.sdl_probe_options.probe_sparse_code_nnz, dtype=torch.uint32 )
        self.register_buffer("probe_sparse_code_nnz", probe_sparse_code_nnz )

        sparse_code_probe = self.get_initial_weights()
        self.register_parameter("sparse_code_probe", torch.nn.Parameter(sparse_code_probe))

        self.build_optimizer()

    def get_dictionary(self):
        dictionary_matrix = torch.tensor( self.options.experimental.sdl_probe_options.d_mat, dtype=torch.complex64 )
        dictionary_matrix_pinv = torch.tensor( self.options.experimental.sdl_probe_options.d_mat_pinv, dtype=torch.complex64 )
        dictionary_matrix_H = torch.tensor( self.options.experimental.sdl_probe_options.d_mat_conj_transpose, dtype=torch.complex64 )
        return dictionary_matrix, dictionary_matrix_pinv, dictionary_matrix_H

    def get_initial_weights(self):
        probe_vec = torch.reshape( self.data, ( self.data.shape[1], self.data.shape[2] * self.data.shape[3] ))
        probe_vec = torch.swapaxes( probe_vec, 0, -1)
        sparse_code_probe = self.dictionary_matrix_pinv @ probe_vec
        return sparse_code_probe

    def generate(self):
        """Generate the probe using the sparse code, and set the
        generated probe to self.data.
        
        Returns
        -------
        Tensor
            A (n_opr_modes, n_modes, h, w) tensor giving the generated probe.
        """
        probe_vec = self.dictionary_matrix @ self.sparse_code_probe
        probe_vec = torch.swapaxes( probe_vec, 0, -1)
        probe = torch.reshape( probe_vec, ( self.data.shape[1], self.data.shape[2], self.data.shape[3] ))[ None, ... ]
        self.tensor.data = torch.stack([probe.real, probe.imag], dim=-1)
        return probe
    
    def build_optimizer(self):
        if self.optimizable and self.optimizer_class is None:
            raise ValueError(
                "Parameter {} is optimizable but no optimizer is specified.".format(self.name)
            )
        if self.optimizable:
            self.optimizer = self.optimizer_class([self.sparse_code_probe], **self.optimizer_params)

    def set_sparse_code(self, data):
        self.sparse_code_probe.data = data


class DIPProbe(Probe):
    
    options: "api.options.ad_ptychography.AutodiffPtychographyProbeOptions"
    representation: ProbeRepresentation = ProbeRepresentation.DIP
    
    def __init__(
        self,
        name: str = "probe",
        options: "api.options.ad_ptychography.AutodiffPtychographyProbeOptions" = None,
        *args,
        **kwargs
    ) -> None:
        """Deep image prior object.

        Parameters
        ----------
        name : str, optional
            The name of the object.
        options : api.options.base.ObjectOptions, optional
            The options for the object.
        """
        super().__init__(name=name, options=options, data_as_parameter=False, *args, **kwargs)
        self.model = None
        self.dip_output_magnitude = None
        self.dip_output_phase = None
        
        self.build_model()
        self.build_dip_optimizer()
        
        # `self.tensor` is used to hold the object generated by the DIP model and
        # is not trainable.
        self.tensor.requires_grad_(False)
        
        nn_input = self.get_nn_input()
        self.register_buffer("nn_input", nn_input)
        
        self.initial_data = None
        if self.options.experimental.deep_image_prior_options.residual_generation:
            self.initial_data = self.data.clone()            
        
    def build_model(self):
        if not self.options.experimental.deep_image_prior_options.enabled:
            return
        model_class = maps.get_nn_model_by_enum(self.options.experimental.deep_image_prior_options.model)
        self.model = model_class(
            **self.options.experimental.deep_image_prior_options.model_params
        )
        
    def build_dip_optimizer(self):
        if self.optimizable and self.optimizer_class is None:
            raise ValueError(
                "Parameter {} is optimizable but no optimizer is specified.".format(self.name)
            )
        if self.optimizable:
            self.optimizer = self.optimizer_class(self.model.parameters(), **self.optimizer_params)

    def get_nn_input(self):
        z = torch.rand(
            [self.n_opr_modes * self.n_modes, 
             self.options.experimental.deep_image_prior_options.net_input_channels, 
             *self.lateral_shape], 
        ) * 0.1
        return z

    def generate(self) -> Tensor:
        """Generate the probe using the deep image prior model, and set the
        generated probe to self.data.
        
        Returns
        -------
        Tensor
            The generated probe.
        """
        if self.model is None:
            raise ValueError("Model is not built.")
        p = self.model(self.nn_input)
        
        p, mag, phase = self.process_net_output(p)
        
        with torch.no_grad():
            self.dip_output_magnitude = mag.clone()
            self.dip_output_phase = phase.clone()
        
        if self.options.experimental.deep_image_prior_options.residual_generation:
            init_data = torch.stack([self.initial_data.real, self.initial_data.imag], dim=-1)
            p = p + init_data
        self.tensor.data = p
        return self.data

    def process_net_output(self, p):
        """Apply constraints to the DIP network's output and transform the tensor.

        Parameters
        ----------
        o : Tensor | tuple[Tensor, Tensor]
            The output of the DIP network. It should either be a [n_modes * n_opr_modes, 2, h, w] 
            tensor with the channels giving the magnitude and phase of the probe,
            or a tuple of two [n_modes * n_opr_modes, h, w] tensors giving the magnitude and phase
            of the probe.
            
        Returns
        -------
        Tensor
            A [n_opr_modes, n_modes, h, w, 2] tensor representing the real and imaginary parts 
            of the probe.
        Tensor
            The magnitude of the probe.
        Tensor
            The phase of the probe.
        """
        if isinstance(p, (list, tuple)):
            mag, phase = p
            mag = mag[:, 0]
            phase = phase[:, 0]
        else:
            mag = p[:, 0]
            phase = p[:, 1]
            
        expected_phase_shape = (self.n_opr_modes * self.n_modes, *self.lateral_shape)
        if tuple(phase.shape) != expected_phase_shape:
            logger.warning(
                "Shape of phase output by NN is expected to be "
                f"(n_slices, h, w) = {expected_phase_shape} but got {phase.shape}. "
                "Padding/cropping generated object to match expected shape."
            )
            mag_resized, phase_resized = [], []
            for i_img in range(expected_phase_shape[0]):
                mag_resized.append(ip.central_crop_or_pad(mag[i_img], expected_phase_shape[1:]))
                phase_resized.append(ip.central_crop_or_pad(phase[i_img], expected_phase_shape[1:]))
            mag = torch.stack(mag_resized)
            phase = torch.stack(phase_resized)
        
        p_complex = mag * torch.exp(1j * phase)
        p = torch.stack([p_complex.real, p_complex.imag], dim=-1)
        p = p.reshape([*self.shape, 2])
        return p, mag, phase
